"""
    Companion stack manager
"""
import logging
import typing
from typing import Dict, List, Optional

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError, NoCredentialsError, NoRegionError

from samcli.commands.exceptions import CredentialsError, RegionError
from samcli.lib.bootstrap.companion_stack.companion_stack_builder import CompanionStackBuilder
from samcli.lib.bootstrap.companion_stack.data_types import CompanionStack, ECRRepo
from samcli.lib.package.artifact_exporter import mktempfile
from samcli.lib.package.s3_uploader import S3Uploader
from samcli.lib.providers.sam_function_provider import SamFunctionProvider
from samcli.lib.providers.sam_stack_provider import SamLocalStackProvider
from samcli.lib.utils.packagetype import IMAGE

# pylint: disable=E0401
if typing.TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_cloudformation.client import CloudFormationClient
    from mypy_boto3_cloudformation.type_defs import WaiterConfigTypeDef
    from mypy_boto3_s3.client import S3Client

LOG = logging.getLogger(__name__)


class CompanionStackManager:
    """
    Manager class for a companion stack
    Used to create/update the remote stack
    """

    _companion_stack: CompanionStack
    _builder: CompanionStackBuilder
    _boto_config: Config
    _update_stack_waiter_config: "WaiterConfigTypeDef"
    _delete_stack_waiter_config: "WaiterConfigTypeDef"
    _s3_bucket: str
    _s3_prefix: str
    _cfn_client: "CloudFormationClient"
    _s3_client: "S3Client"

    def __init__(self, stack_name, region, s3_bucket, s3_prefix):
        self._companion_stack = CompanionStack(stack_name)
        self._builder = CompanionStackBuilder(self._companion_stack)
        self._boto_config = Config(region_name=region if region else None)
        self._update_stack_waiter_config = {"Delay": 10, "MaxAttempts": 120}
        self._delete_stack_waiter_config = {"Delay": 10, "MaxAttempts": 120}
        self._s3_bucket = s3_bucket
        self._s3_prefix = s3_prefix
        try:
            self._cfn_client = boto3.client("cloudformation", config=self._boto_config)
            self._ecr_client = boto3.client("ecr", config=self._boto_config)
            self._s3_client = boto3.client("s3", config=self._boto_config)
            self._account_id = boto3.client("sts").get_caller_identity().get("Account")
            self._region_name = self._cfn_client.meta.region_name
        except NoCredentialsError as ex:
            raise CredentialsError(
                "Error Setting Up Managed Stack Client: Unable to resolve "
                "credentials for the AWS SDK for Python client. "
                "Please see their documentation for options to pass in credentials: "
                "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html"
            ) from ex
        except NoRegionError as ex:
            raise RegionError(
                "Error Setting Up Managed Stack Client: Unable to resolve a region. "
                "Please provide a region via the --region parameter or by the AWS_REGION environment variable."
            ) from ex

    def set_functions(
        self, function_logical_ids: List[str], image_repositories: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Sets functions that need to have ECR repos created

        Parameters
        ----------
        function_logical_ids: List[str]
            Function logical IDs that need to have ECR repos created
        image_repositories: Optional[Dict[str, str]]
            Optional image repository mapping. Functions with non-auto-ecr URIs
            will be ignored.
        """
        self._builder.clear_functions()
        if image_repositories is None:
            image_repositories = dict()
        for function_logical_id in function_logical_ids:
            if function_logical_id not in image_repositories or self.is_repo_uri(
                image_repositories.get(function_logical_id), function_logical_id
            ):
                self._builder.add_function(function_logical_id)

    def update_companion_stack(self) -> None:
        """
        Blocking call to create or update the companion stack based on current functions
        Companion stack template will be updated to the s3 bucket first before deployment
        """
        if not self._builder.repo_mapping:
            return

        stack_name = self._companion_stack.stack_name
        template = self._builder.build()

        with mktempfile() as temporary_file:
            temporary_file.write(template)
            temporary_file.flush()

            s3_uploader = S3Uploader(
                self._s3_client, bucket_name=self._s3_bucket, prefix=self._s3_prefix, no_progressbar=True
            )
            # TemplateUrl property requires S3 URL to be in path-style format
            parts = S3Uploader.parse_s3_url(
                s3_uploader.upload_with_dedup(temporary_file.name, "template"), version_property="Version"
            )

        template_url = s3_uploader.to_path_style_s3_url(parts["Key"], parts.get("Version", None))

        exists = self.does_companion_stack_exist()
        if exists:
            self._cfn_client.update_stack(
                StackName=stack_name, TemplateURL=template_url, Capabilities=["CAPABILITY_AUTO_EXPAND"]
            )
            update_waiter = self._cfn_client.get_waiter("stack_update_complete")
            update_waiter.wait(StackName=stack_name, WaiterConfig=self._update_stack_waiter_config)
        else:
            self._cfn_client.create_stack(
                StackName=stack_name, TemplateURL=template_url, Capabilities=["CAPABILITY_AUTO_EXPAND"]
            )
            create_waiter = self._cfn_client.get_waiter("stack_create_complete")
            create_waiter.wait(StackName=stack_name, WaiterConfig=self._update_stack_waiter_config)

    def _delete_companion_stack(self) -> None:
        """
        Blocking call to delete the companion stack
        """
        stack_name = self._companion_stack.stack_name
        waiter = self._cfn_client.get_waiter("stack_delete_complete")
        self._cfn_client.delete_stack(StackName=stack_name)
        waiter.wait(StackName=stack_name, WaiterConfig=self._delete_stack_waiter_config)

    def list_deployed_repos(self) -> List[ECRRepo]:
        """
        List deployed ECR repos for this companion stack
        Not using create_change_set as it is slow.

        Returns
        -------
        List[ECRRepo]
            List of ECR repos deployed for this companion stack
            Returns empty list if companion stack does not exist
        """
        if not self.does_companion_stack_exist():
            return []
        repos: List[ECRRepo] = list()
        stack = boto3.resource("cloudformation", config=self._boto_config).Stack(self._companion_stack.stack_name)
        for resource in stack.resource_summaries.all():
            if resource.resource_type == "AWS::ECR::Repository":
                repos.append(
                    ECRRepo(logical_id=resource.logical_resource_id, physical_id=resource.physical_resource_id)
                )
        return repos

    def get_unreferenced_repos(self) -> List[ECRRepo]:
        """
        List deployed ECR repos that is not referenced by current list of functions

        Returns
        -------
        List[ECRRepo]
            List of deployed ECR repos that is not referenced by current list of functions
            Returns empty list if companion stack does not exist
        """
        if not self.does_companion_stack_exist():
            return []
        deployed_repos: List[ECRRepo] = self.list_deployed_repos()
        current_mapping = self._builder.repo_mapping

        unreferenced_repos: List[ECRRepo] = list()
        for deployed_repo in deployed_repos:
            for _, current_repo in current_mapping.items():
                if current_repo.logical_id == deployed_repo.logical_id:
                    break
            else:
                unreferenced_repos.append(deployed_repo)
        return unreferenced_repos

    def delete_unreferenced_repos(self) -> None:
        """
        Blocking call to delete all deployed ECR repos that are unreferenced by a function
        If repo does not exist, this will simply skip it.
        """
        repos = self.get_unreferenced_repos()
        for repo in repos:
            try:
                self._ecr_client.delete_repository(repositoryName=repo.physical_id, force=True)
            except self._ecr_client.exceptions.RepositoryNotFoundException:
                LOG.debug("Image repo [%s] not found in companion stack. Skipping deletion.", repo.physical_id)

    def sync_repos(self) -> None:
        """
        Blocking call to sync companion stack with the following actions
        Creates the stack if it does not exist, and updates it if it does.
        Deletes unreferenced repos if they exist.
        Deletes companion stack if there isn't any repo left.
        """
        has_repo = bool(self.get_repository_mapping())
        if self.does_companion_stack_exist():
            self.delete_unreferenced_repos()
            if has_repo:
                self.update_companion_stack()
            else:
                self._delete_companion_stack()
        elif has_repo:
            self.update_companion_stack()

    def does_companion_stack_exist(self) -> bool:
        """
        Does companion stack exist

        Returns
        -------
        bool
            Returns True if companion stack exists
        """
        try:
            self._cfn_client.describe_stacks(StackName=self._companion_stack.stack_name)
            return True
        except ClientError as e:
            error_message = e.response.get("Error", {}).get("Message")
            if error_message == f"Stack with id {self._companion_stack.stack_name} does not exist":
                return False
            raise e

    def get_repository_mapping(self) -> Dict[str, str]:
        """
        Get current function to repo mapping

        Returns
        -------
        Dict[str, str]
            Dictionary with key as function logical ID and value as ECR repo URI.
        """
        return dict((k, self.get_repo_uri(v)) for (k, v) in self._builder.repo_mapping.items())

    def get_repo_uri(self, repo: ECRRepo) -> str:
        """
        Get repo URI for a ECR repo

        Parameters
        ----------
        repo: ECRRepo

        Returns
        -------
        str
            ECR repo URI based on account ID and region.
        """
        return repo.get_repo_uri(self._account_id, self._region_name)

    def is_repo_uri(self, repo_uri: Optional[str], function_logical_id: str) -> bool:
        """
        Check whether repo URI is a companion stack repo

        Parameters
        ----------
        repo_uri: str
            Repo URI to be checked.

        function_logical_id: str
            Function logical ID associated with the image repo.

        Returns
        -------
        bool
            Returns True if repo_uri is a companion stack repo.
        """
        return repo_uri == self.get_repo_uri(ECRRepo(self._companion_stack, function_logical_id))


def sync_ecr_stack(
    template_file: str, stack_name: str, region: str, s3_bucket: str, s3_prefix: str, image_repositories: Dict[str, str]
) -> Dict[str, str]:
    """Blocking call to sync local functions with ECR Companion Stack

    Parameters
    ----------
    template_file : str
        Template file path.
    stack_name : str
        Stack name
    region : str
        AWS region
    s3_bucket : str
        S3 bucket
    s3_prefix : str
        S3 prefix for the bucket
    image_repositories : Dict[str, str]
        Mapping between function logical ID and ECR URI

    Returns
    -------
    Dict[str, str]
        Updated mapping of image_repositories. Auto ECR URIs are added
        for Functions without a repo specified.
    """
    image_repositories = image_repositories.copy() if image_repositories else {}
    manager = CompanionStackManager(stack_name, region, s3_bucket, s3_prefix)

    stacks = SamLocalStackProvider.get_stacks(template_file)[0]
    function_provider = SamFunctionProvider(stacks, ignore_code_extraction_warnings=True)
    function_logical_ids = [
        function.full_path for function in function_provider.get_all() if function.packagetype == IMAGE
    ]
    manager.set_functions(function_logical_ids, image_repositories)
    image_repositories.update(manager.get_repository_mapping())
    manager.sync_repos()
    return image_repositories



"""
Configures a logger
"""
import logging

SAM_CLI_FORMATTER = logging.Formatter("%(message)s")
SAM_CLI_FORMATTER_WITH_TIMESTAMP = logging.Formatter("%(asctime)s | %(message)s")

SAM_CLI_LOGGER_NAME = "samcli"
LAMBDA_BULDERS_LOGGER_NAME = "aws_lambda_builders"


class SamCliLogger:
    @staticmethod
    def configure_logger(logger, formatter, level):
        """
        Configure a Logger with the level provided and also the first handler's formatter.
        If there is no handler in the logger, a new StreamHandler will be added.

        Parameters
        ----------
        logger logging.getLogger
            Logger to configure
        formatter logging.formatter
            Formatter for the logger
        """
        handlers = logger.handlers
        if handlers:
            log_stream_handler = handlers[0]
        else:
            log_stream_handler = logging.StreamHandler()
            logger.addHandler(log_stream_handler)

        log_stream_handler.setLevel(logging.DEBUG)
        log_stream_handler.setFormatter(formatter)

        logger.setLevel(level)
        logger.propagate = False

    @staticmethod
    def configure_null_logger(logger):
        """
        Configure a Logger with a NullHandler

        Useful for libraries that do not follow:
            https://docs.python.org/3.9/howto/logging.html#configuring-logging-for-a-library

        Parameters
        ----------
        logger logging.getLogger
            Logger to configure
        """
        logger.propagate = False
        logger.addHandler(logging.NullHandler())



"""
Connects the CLI with Local API Gateway service.
"""

import logging
import os

from samcli.commands.local.lib.exceptions import NoApisDefined
from samcli.lib.providers.api_provider import ApiProvider
from samcli.local.apigw.local_apigw_service import LocalApigwService

LOG = logging.getLogger(__name__)


class LocalApiService:
    """
    Implementation of Local API service that is capable of serving API defined in a configuration file that invoke a
    Lambda function.
    """

    def __init__(self, lambda_invoke_context, port, host, static_dir):
        """
        Initialize the local API service.

        :param samcli.commands.local.cli_common.invoke_context.InvokeContext lambda_invoke_context: Context object
            that can help with Lambda invocation
        :param int port: Port to listen on
        :param string host: Local hostname or IP address to bind to
        :param string static_dir: Optional, directory from which static files will be mounted
        """

        self.port = port
        self.host = host
        self.static_dir = static_dir

        self.cwd = lambda_invoke_context.get_cwd()
        self.api_provider = ApiProvider(lambda_invoke_context.stacks, cwd=self.cwd)
        self.lambda_runner = lambda_invoke_context.local_lambda_runner
        self.stderr_stream = lambda_invoke_context.stderr

    def start(self):
        """
        Creates and starts the local API Gateway service. This method will block until the service is stopped
        manually using an interrupt. After the service is started, callers can make HTTP requests to the endpoint
        to invoke the Lambda function and receive a response.

        NOTE: This is a blocking call that will not return until the thread is interrupted with SIGINT/SIGTERM
        """

        if not self.api_provider.api.routes:
            raise NoApisDefined("No APIs available in template")

        static_dir_path = self._make_static_dir_path(self.cwd, self.static_dir)

        # We care about passing only stderr to the Service and not stdout because stdout from Docker container
        # contains the response to the API which is sent out as HTTP response. Only stderr needs to be printed
        # to the console or a log file. stderr from Docker container contains runtime logs and output of print
        # statements from the Lambda function
        service = LocalApigwService(
            api=self.api_provider.api,
            lambda_runner=self.lambda_runner,
            static_dir=static_dir_path,
            port=self.port,
            host=self.host,
            stderr=self.stderr_stream,
        )

        service.create()

        # Print out the list of routes that will be mounted
        self._print_routes(self.api_provider.api.routes, self.host, self.port)
        LOG.info(
            "You can now browse to the above endpoints to invoke your functions. "
            "You do not need to restart/reload SAM CLI while working on your functions, "
            "changes will be reflected instantly/automatically. If you used sam build before "
            "running local commands, you will need to re-run sam build for the changes "
            "to be picked up. You only need to restart SAM CLI if you update your AWS SAM template"
        )

        service.run()

    @staticmethod
    def _print_routes(routes, host, port):
        """
        Helper method to print the APIs that will be mounted. This method is purely for printing purposes.
        This method takes in a list of Route Configurations and prints out the Routes grouped by path.
        Grouping routes by Function Name + Path is the bulk of the logic.

        Example output:
            Mounting Product at http://127.0.0.1:3000/path1/bar [GET, POST, DELETE]
            Mounting Product at http://127.0.0.1:3000/path2/bar [HEAD]

        :param list(Route) routes:
            List of routes grouped by the same function_name and path
        :param string host:
            Host name where the service is running
        :param int port:
            Port number where the service is running
        :returns list(string):
            List of lines that were printed to the console. Helps with testing
        """

        print_lines = []
        for route in routes:
            methods_str = "[{}]".format(", ".join(route.methods))
            output = "Mounting {} at http://{}:{}{} {}".format(route.function_name, host, port, route.path, methods_str)
            print_lines.append(output)

            LOG.info(output)

        return print_lines

    @staticmethod
    def _make_static_dir_path(cwd, static_dir):
        """
        This method returns the path to the directory where static files are to be served from. If static_dir is a
        relative path, then it is resolved to be relative to the current working directory. If no static directory is
        provided, or if the resolved directory does not exist, this method will return None

        :param string cwd: Current working directory relative to which we will resolve the static directory
        :param string static_dir: Path to the static directory
        :return string: Path to the static directory, if it exists. None, otherwise
        """
        if not static_dir:
            return None

        static_dir_path = os.path.join(cwd, static_dir)
        if os.path.exists(static_dir_path):
            LOG.info("Mounting static files from %s at /", static_dir_path)
            return static_dir_path

        return None



"""
Use Terraform plan to link resources together
e.g. linking layers to functions
"""
import logging
import re
from typing import Dict, List, Optional, Union

from samcli.hook_packages.terraform.hooks.prepare.exceptions import (
    InvalidResourceLinkingException,
    LocalVariablesLinkingLimitationException,
    OneLambdaLayerLinkingLimitationException,
)
from samcli.hook_packages.terraform.hooks.prepare.types import (
    ConstantValue,
    Expression,
    References,
    ResolvedReference,
    TFModule,
    TFResource,
)
from samcli.hook_packages.terraform.lib.utils import build_cfn_logical_id

LAMBDA_LAYER_RESOURCE_ADDRESS_PREFIX = "aws_lambda_layer_version."
TERRAFORM_LOCAL_VARIABLES_ADDRESS_PREFIX = "local."
DATA_RESOURCE_ADDRESS_PREFIX = "data."

LOG = logging.getLogger(__name__)
COMPILED_REGULAR_EXPRESSION = re.compile(r"\[[^\[\]]*\]")


def _build_module(
    module_name: Optional[str],
    module_configuration: Dict,
    input_variables: Dict[str, Expression],
    parent_module_address: Optional[str],
) -> TFModule:
    """
    Builds and returns a TFModule

    Parameters
    ==========
    module_name: Optional[str]
        The module's name, if any
    module_configuration: Dict
        The module object from the terraform configuration
    input_variables: Dict[str, Expression]
        The input variables sent into the module
    parent_module_address: Optional[str]
        The module's parent address, if any

    Returns
    =======
    TFModule
        The constructed TFModule
    """
    module = TFModule(None, None, {}, {}, {}, {})

    module.full_address = _build_module_full_address(module_name, parent_module_address)
    LOG.debug("Parsing module:` %s", module.full_address or "root")

    if not module_configuration:
        raise InvalidResourceLinkingException(f"No module configuration for module: {module.full_address or 'root'}")

    LOG.debug("Parsing module variables")
    module.variables = _build_module_variables_from_configuration(module_configuration, input_variables)

    LOG.debug("Parsing module resources")
    module.resources = _build_module_resources_from_configuration(module_configuration, module)

    LOG.debug("Parsing module outputs")
    module.outputs = _build_module_outputs_from_configuration(module_configuration)

    LOG.debug("Parsing module calls")
    module.child_modules = _build_child_modules_from_configuration(module_configuration, module)

    return module


def _build_module_full_address(module_name: Optional[str], parent_module_address: Optional[str]) -> Optional[str]:
    """
    Returns the full address of a module, depending on whether it has a module name and a parent module address.

    Parameters
    ==========
    module_name: Optional[str]
        The module's name, if any
    parent_module_address: Optional[str]
        The module's parent address, if any

    Returns
    =======
    Optional[str]
        Returns None if no module_name is provided (e.g. root module).
        Returns module.<module_name> if a module_name is provided
        Returns <parent_module_address>.module.<module_name> if both module_name and
            parent_module_address are provided
    """
    full_address = None
    if module_name:
        full_address = f"module.{module_name}"
        if parent_module_address:
            full_address = f"{parent_module_address}.{full_address}"

    return full_address


def _build_module_variables_from_configuration(
    module_configuration: Dict, input_variables: Dict[str, Expression]
) -> Dict[str, Expression]:
    """
    Builds and returns module variables as Expressions using a module terraform configuration

    Parameters
    ==========
    module_configuration: dict
        The module object from the terraform configuration
    input_variables: Dict[str, Expression]
        The input variables sent into the module to override default variables

    Returns
    =======
    Dict[str, Expression]
        Dictionary with the variable names as keys and parsed Expression as values.
    """
    module_variables: Dict[str, Expression] = {}

    default_variables = module_configuration.get("variables", {})
    for variable_name, variable_value in default_variables.items():
        module_variables[variable_name] = ConstantValue(variable_value.get("default"))
    module_variables.update(input_variables)

    return module_variables


def _build_module_resources_from_configuration(module_configuration: Dict, module: TFModule) -> Dict[str, TFResource]:
    """
    Builds and returns module TFResources using a module terraform configuration

    Parameters
    ==========
    module_configuration: dict
        The module object from the terraform configuration
    module: TFModule
        The TFModule whose resources we're parsing

    Returns
    =======
    Dict[TFResource]
        Dictionary of TFResource for the parsed resources from the config, and the key is the resource address
    """
    module_resources = {}

    config_resources = module_configuration.get("resources", [])
    for config_resource in config_resources:
        resource_attributes: Dict[str, Expression] = {}

        expressions = config_resource.get("expressions", {})
        for expression_name, expression_value in expressions.items():
            # we do not process the attributes of type dictionary
            # Todo add dictionary type attributes post beta
            if isinstance(expression_value, list):
                LOG.debug("Skip processing the attribute %s as its value is a map.", expression_name)
                continue

            parsed_expression = _build_expression_from_configuration(expression_value)
            if parsed_expression:
                resource_attributes[expression_name] = parsed_expression

        resource_address = config_resource.get("address")
        resource_type = config_resource.get("type")
        module_resources[resource_address] = TFResource(resource_address, resource_type, module, resource_attributes)

    return module_resources


def _build_module_outputs_from_configuration(module_configuration: Dict) -> Dict[str, Expression]:
    """
    Builds and returns module outputs as Expressions using a module terraform configuration

    Parameters
    ==========
    module_configuration: dict
        The module object from the terraform configuration

    Returns
    =======
    Dict[str, Expression]
        Dictionary with the output names as keys and parsed Expression as values.
    """
    module_outputs = {}

    config_outputs = module_configuration.get("outputs", {})
    for output_name, output_value in config_outputs.items():
        expression = output_value.get("expression", {})
        parsed_expression = _build_expression_from_configuration(expression)
        if parsed_expression:
            module_outputs[output_name] = parsed_expression

    return module_outputs


def _build_child_modules_from_configuration(module_configuration: Dict, module: TFModule) -> Dict[str, TFModule]:
    """
    Builds and returns child TFModules using a module terraform configuration

    Parameters
    ==========
    module_configuration: dict
        The module object from the terraform configuration
    module: TFModule
        The TFModule whose child modules we're building

    Returns
    =======
    Dict[str, TFModule]
        Dictionary with the module names as keys and parsed TFModule as values.
    """
    child_modules = {}

    module_calls = module_configuration.get("module_calls", {})
    for module_call_name, module_call_value in module_calls.items():
        module_call_input_variables: Dict[str, Expression] = {}

        expressions = module_call_value.get("expressions", {})
        for expression_name, expression_value in expressions.items():
            parsed_expression = _build_expression_from_configuration(expression_value)
            if parsed_expression:
                module_call_input_variables[expression_name] = parsed_expression

        module_call_module_config = module_call_value.get("module", {})
        module_call_built_module = _build_module(
            module_call_name, module_call_module_config, module_call_input_variables, module.full_address
        )

        module_call_built_module.parent_module = module
        child_modules[module_call_name] = module_call_built_module

    return child_modules


def _build_expression_from_configuration(expression_configuration: Dict) -> Optional[Expression]:
    """
    Parses an Expression from an expression terraform configuration.

    Parameters
    ==========
    expression_configuration: dict
        The expression object from the terraform configuration

    Returns
    =======
    Expression
        The parsed expression
    """
    constant_value = expression_configuration.get("constant_value")
    references = expression_configuration.get("references")

    parsed_expression: Optional[Expression] = None

    if constant_value is not None:
        parsed_expression = ConstantValue(constant_value)
    elif references is not None:
        parsed_expression = References(references)

    return parsed_expression


def _clean_references_list(references: List[str]) -> List[str]:
    """
    Return a new copy of the complete references list.

    e.g. given a list of references like
    [
        'aws_lambda_layer_version.layer1[0].arn',
        'aws_lambda_layer_version.layer1[0]',
        'aws_lambda_layer_version.layer1',
    ]
    We want only the first complete reference ('aws_lambda_layer_version.layer1[0].arn')

    Parameters
    ----------
    references: List[str]
        A list of reference strings

    Returns
    -------
    List[str]
        A copy of a cleaned list of reference strings
    """
    cleaned_references = []
    copied_references = sorted(references, reverse=True)
    if not references:
        return []
    cleaned_references.append(copied_references[0])
    for i in range(1, len(copied_references)):
        if not cleaned_references[-1].startswith(copied_references[i]):
            cleaned_references.append(copied_references[i])
    return cleaned_references


def _get_configuration_address(address: str) -> str:
    """
    Cleans all addresses of indices and returns a clean address

    Parameters
    ----------
    address : str
        The address to clean

    Returns
    -------
    str
        The address clean of indices
    """
    return COMPILED_REGULAR_EXPRESSION.sub("", address)


def _resolve_module_output(module: TFModule, output_name: str) -> List[Union[ConstantValue, ResolvedReference]]:
    """
    Resolves any references in the output section of the module

    Parameters
    ----------
    module : Module
        The module with outputs to search
    output_name : str
        The value to resolve

    Returns
    -------
    List[Union[ConstantValue, ResolvedReference]]
        A list of resolved values
    """
    results: List[Union[ConstantValue, ResolvedReference]] = []

    output = module.outputs.get(output_name)

    if not output:
        raise InvalidResourceLinkingException(f"Output {output_name} was not found in module {module.full_address}")

    output_value = output.value

    LOG.debug("Resolving output {%s} for module {%s}", output_name, module.full_address)

    if isinstance(output, ConstantValue):
        LOG.debug(
            "Resolved constant value {%s} for module {%s} for output {%s}",
            output.value,
            module.full_address,
            output_name,
        )

        results.append(output)
    elif isinstance(output, References):
        LOG.debug("Found references for module {%s} for output {%s}", module.full_address, output_name)

        cleaned_references = _clean_references_list(output_value)

        for reference in cleaned_references:
            if reference.startswith("var."):
                LOG.debug(
                    "Resolving variable reference {%s} for module {%s} for output {%s}",
                    reference,
                    module.full_address,
                    output_name,
                )

                stripped_reference = _get_configuration_address(reference[reference.find(".") + 1 :])
                results += _resolve_module_variable(module, stripped_reference)
            elif reference.startswith("module."):
                LOG.debug(
                    "Resolving module reference {%s} for module {%s} for output {%s}",
                    reference,
                    module.full_address,
                    output_name,
                )

                # validate that the reference is in the format: module.name.output
                if re.fullmatch(r"module(?:\.[^\.]+){2}", reference) is None:
                    raise InvalidResourceLinkingException(
                        f"Module {module.full_address} contains an invalid reference {reference}"
                    )

                # module.bbb.ccc => bbb
                module_name = reference[reference.find(".") + 1 : reference.rfind(".")]
                # module.bbb.ccc => ccc
                output_name = reference[reference.rfind(".") + 1 :]

                stripped_reference = _get_configuration_address(module_name)

                if not module.child_modules:
                    raise InvalidResourceLinkingException(
                        f"Module {module.full_address} does not have child modules defined"
                    )

                child_module = module.child_modules.get(stripped_reference)

                if not child_module:
                    raise InvalidResourceLinkingException(
                        f"Module {module.full_address} does not have {stripped_reference} as a child module"
                    )

                results += _resolve_module_output(child_module, output_name)
            else:
                LOG.debug(
                    "Resolved reference {%s} for module {%s} for output {%s}",
                    reference,
                    module.full_address,
                    output_name,
                )

                results.append(ResolvedReference(reference, module.full_address))

    return results


def _resolve_module_variable(module: TFModule, variable_name: str) -> List[Union[ConstantValue, ResolvedReference]]:
    # return a list of the values that resolve the passed variable
    # name in the input module.
    results: List[Union[ConstantValue, ResolvedReference]] = []

    LOG.debug("Resolving module variable for module (%s) and variable (%s)", module.module_name, variable_name)

    var_value = module.variables.get(variable_name)

    if not var_value:
        raise InvalidResourceLinkingException(
            message=f"The variable {variable_name} could not be found in module {module.module_name}."
        )

    # check the possible constant value for this variable
    if isinstance(var_value, ConstantValue) and var_value is not None:
        LOG.debug("Found a constant value (%s) in module (%s)", var_value.value, module.module_name)
        results.append(ConstantValue(var_value.value))

    # check the possible references value for this variable
    if isinstance(var_value, References) and var_value is not None:
        LOG.debug("Found references (%s) in module (%s)", var_value.value, module.module_name)
        cleaned_references = _clean_references_list(var_value.value)
        for reference in cleaned_references:
            LOG.debug("Resolving reference: %s", reference)
            # refer to a variable passed to this module from its parent module
            if reference.startswith("var."):
                config_var_name = _get_configuration_address(reference[len("var.") :])
                if module.parent_module:
                    results += _resolve_module_variable(module.parent_module, config_var_name)
            # refer to another module output. This module will be defined in the same level as this module
            elif reference.startswith("module."):
                module_name = reference[reference.find(".") + 1 : reference.rfind(".")]
                config_module_name = _get_configuration_address(module_name)
                output_name = reference[reference.rfind(".") + 1 :]
                if (
                    module.parent_module
                    and module.parent_module.child_modules
                    and module.parent_module.child_modules.get(config_module_name)
                ):
                    # using .get() gives us Optional[TFModule], if conditional already validates child module exists
                    # access list directly instead
                    child_module = module.parent_module.child_modules[config_module_name]
                    results += _resolve_module_output(child_module, output_name)
                else:
                    raise InvalidResourceLinkingException(f"Couldn't find child module {config_module_name}.")
            # this means either a resource, data source, or local.variables.
            elif module.parent_module:
                results.append(ResolvedReference(reference, module.parent_module.full_address))
            else:
                raise InvalidResourceLinkingException("Resource linking entered an invalid state.")

    return results


def _resolve_resource_attribute(
    resource: TFResource, attribute_name: str
) -> List[Union[ConstantValue, ResolvedReference]]:
    """
    Return a list of the values that resolve the passed attribute name in the input terraform resource configuration

    Parameters
    ----------
    resource: TFResource
        A terraform resource

    attribute_name: str
        The attribute name that needs to be resolved.

    Returns
    -------
    List[Union[ConstantValue, ResolvedReference]]
        A list of combination of constant values and/or references to other terraform resources attributes.
    """

    results: List[Union[ConstantValue, ResolvedReference]] = []
    LOG.debug(
        "Resolving resource attribute for resource (%s) and attribute (%s)", resource.full_address, attribute_name
    )

    attribute_value = resource.attributes.get(attribute_name)
    if attribute_value is None:
        LOG.debug("The value of the attribute %s is None for resource %s", attribute_name, resource.full_address)
        return results

    if not isinstance(attribute_value, ConstantValue) and not isinstance(attribute_value, References):
        raise InvalidResourceLinkingException(
            message=f"The attribute {attribute_name} has unexpected type in resource {resource.full_address}."
        )

    # check the possible constant value for this attribute
    if isinstance(attribute_value, ConstantValue):
        LOG.debug(
            "Found a constant value (%s) for attribute (%s) in resource (%s)",
            attribute_value.value,
            attribute_name,
            resource.full_address,
        )
        results.append(ConstantValue(attribute_value.value))
        return results

    # resolve the attribute reference value
    LOG.debug(
        "Found references (%s) for attribute (%s) in resource (%s)",
        attribute_value.value,
        attribute_name,
        resource.full_address,
    )
    cleaned_references = _clean_references_list(attribute_value.value)

    for reference in cleaned_references:
        # refer to a variable passed to this resource module from its parent module
        if reference.startswith("var."):
            config_var_name = _get_configuration_address(reference[len("var.") :])
            LOG.debug("Traversing a variable reference: %s to variable named %s", reference, config_var_name)
            results += _resolve_module_variable(resource.module, config_var_name)

        # refer to another module output. This module will be defined in the same level as the resource
        elif reference.startswith("module."):
            # validate that the reference is in the format: module.name.output
            if re.fullmatch(r"module(?:\.[^\.]+){2}", reference) is None:
                LOG.debug("Could not traverse the module output reference: %s", reference)
                raise InvalidResourceLinkingException(
                    f"The attribute {attribute_name} in Resource {resource.full_address} has an invalid reference "
                    f"{reference} value"
                )

            module_name = reference[reference.find(".") + 1 : reference.rfind(".")]
            config_module_name = _get_configuration_address(module_name)
            output_name = reference[reference.rfind(".") + 1 :]
            LOG.debug(
                "Traversing the module output reference: %s to the output named %s in module %s",
                reference,
                output_name,
                config_module_name,
            )

            if not resource.module.child_modules or resource.module.child_modules.get(config_module_name) is None:
                raise InvalidResourceLinkingException(
                    f"The input resource {resource.full_address} does not have a parent module, or we could not "
                    f"find the child module {config_module_name}."
                )

            results += _resolve_module_output(resource.module.child_modules[config_module_name], output_name)

        # this means either a resource, data source, or local.variables.
        else:
            results.append(ResolvedReference(reference, resource.module.full_address))
    return results


def _link_lambda_function_to_layer(
    function_tf_resource: TFResource, cfn_functions: List[Dict], tf_layers: Dict[str, Dict]
) -> None:
    """
    Resolve the lambda layer for the input lambda function configuration resource, and then update the equivalent cfn
    lambda functions list.
    The Lambda configuration resource in Terraform can match multiple actual resources in case if it was defined using
    count or for_each pattern.

    Parameters
    ----------
    function_tf_resource: TFResource
        The input lambda function terraform configuration resource

    cfn_functions: List[Dict]
        A list of mapped lambda functions that are equivalent to the input terraform configuration lambda function

    tf_layers: Dict[str, Dict]
        Dictionary of all actual terraform layers resources (not configuration resources). The dictionary's key is the
        calculated logical id for each resource
    """

    LOG.debug(
        "Link function configuration %s layer that has these instances %s.",
        function_tf_resource.full_address,
        cfn_functions,
    )
    resolved_layers = _resolve_resource_attribute(function_tf_resource, "layers")
    LOG.debug("The resolved layers for function %s are %s", function_tf_resource.full_address, resolved_layers)
    layers = _process_resolved_layers(function_tf_resource, resolved_layers, tf_layers)

    # The agreed limitation to support only 1 lambda layer reference.
    if len(layers) > 1:
        LOG.debug(
            "AWS SAM CLI does not support mapping the lambda function %s to more than one layer.",
            function_tf_resource.full_address,
        )
        raise OneLambdaLayerLinkingLimitationException(layers, function_tf_resource.full_address)
    if not layers:
        LOG.debug("There are no layers defined for lambda function %s", function_tf_resource.full_address)
    else:
        _update_mapped_lambda_function_with_resolved_layers(cfn_functions, layers, tf_layers)


def _process_resolved_layers(
    function_tf_resource: TFResource,
    resolved_layers: List[Union[ConstantValue, ResolvedReference]],
    tf_layers: Dict[str, Dict],
) -> List[Dict[str, str]]:
    """
    Process the resolved layers.

    Parameters
    ----------
    function_tf_resource: TFResource
        The input lambda function terraform configuration resource

    resolved_layers: List[Union[ConstantValue, ResolvedReference]]
        The resolved layers to be processed for the input lambda function.

    tf_layers: Dict[str, Dict]
        Dictionary of all actual terraform layers resources (not configuration resources). The dictionary's key is the
        calculated logical id for each resource

    Returns
    --------
    List[Dict[str, str]]:
        The list of layers after processing

    """
    LOG.debug(
        "Map the resolved layers %s to configuration function %s.", resolved_layers, function_tf_resource.full_address
    )
    layers = []
    does_refer_to_constant_values = False
    does_refer_to_data_sources = False
    for resolved_layer in resolved_layers:
        # Skip ConstantValue layers reference, as it will be already handled by terraform plan command.
        if isinstance(resolved_layer, ConstantValue):
            does_refer_to_constant_values = True
        elif isinstance(resolved_layer, ResolvedReference):
            processed_layers = _process_reference_layer_value(function_tf_resource, resolved_layer, tf_layers)
            if not processed_layers:
                does_refer_to_data_sources = True
            layers += processed_layers

    if (does_refer_to_constant_values or does_refer_to_data_sources) and len(layers) > 0:
        LOG.debug(
            "Lambda function %s to is referring to layers using an expression mixing between constant value or data "
            "sources and other resources references. AWS SAM CLI could not determine this lambda functions layers.",
            function_tf_resource.full_address,
        )
        raise OneLambdaLayerLinkingLimitationException(resolved_layers, function_tf_resource.full_address)

    return layers


def _process_reference_layer_value(
    function_tf_resource: TFResource, resolved_layer: ResolvedReference, tf_layers: Dict[str, Dict]
) -> List[Dict[str, str]]:
    """
    Process the layer value of type ResolvedReference.

    Parameters
    ----------
    function_tf_resource: TFResource
        The input lambda function terraform configuration resource

    resolved_layer: ResolvedReference
        The layer reference.

    tf_layers: Dict[str, Dict]
        Dictionary of all actual terraform layers resources (not configuration resources). The dictionary's key is the
        calculated logical id for each resource

    Returns
    -------
    List[Dict[str, str]]
        The resolved layers values that will be used as a value for the mapped CFN function Layers attribute.

    """
    LOG.debug("Process the reference layer %s.", resolved_layer.value)
    # skip processing the data source block, as it should be mapped while executing the terraform plan command.
    if resolved_layer.value.startswith(DATA_RESOURCE_ADDRESS_PREFIX):
        LOG.debug("Skip processing the reference layer %s, as it is referring to a data resource", resolved_layer.value)
        return []

    # resolved reference is a local variable
    if resolved_layer.value.startswith(TERRAFORM_LOCAL_VARIABLES_ADDRESS_PREFIX):
        LOG.debug("AWS SAM CLI could not process the Local variables %s", resolved_layer.value)
        raise LocalVariablesLinkingLimitationException(resolved_layer.value, function_tf_resource.full_address)

    # Valid Layer resource
    if resolved_layer.value.startswith(LAMBDA_LAYER_RESOURCE_ADDRESS_PREFIX):
        LOG.debug("Process the Layer version resource %s", resolved_layer.value)
        if not resolved_layer.value.endswith("arn"):
            LOG.debug("The used property in reference %s is not an ARN property", resolved_layer.value)
            raise InvalidResourceLinkingException(
                f"Could not use the value {resolved_layer.value} as a Layer for lambda function "
                f"{function_tf_resource.full_address}. Lambda Function Layer value should refer to valid "
                f"lambda layer ARN property"
            )

        tf_layer_res_name = resolved_layer.value[len(LAMBDA_LAYER_RESOURCE_ADDRESS_PREFIX) : -len(".arn")]
        if resolved_layer.module_address:
            tf_layer_full_address = (
                f"{resolved_layer.module_address}.{LAMBDA_LAYER_RESOURCE_ADDRESS_PREFIX}" f"{tf_layer_res_name}"
            )
        else:
            tf_layer_full_address = f"{LAMBDA_LAYER_RESOURCE_ADDRESS_PREFIX}{tf_layer_res_name}"
        cfn_layer_logical_id = build_cfn_logical_id(tf_layer_full_address)
        LOG.debug("The logical id of the resource referred by %s is %s", resolved_layer.value, cfn_layer_logical_id)

        # validate that the found layer is in mapped layers resources, which means that it is created.
        # The resource can be defined in the TF plan configuration, but will not be created.
        layers = []
        if cfn_layer_logical_id in tf_layers:
            LOG.debug("The resource referred by %s can be found in the mapped layers resources", resolved_layer.value)
            layers.append({"Ref": cfn_layer_logical_id})
        return layers
    # it means the Lambda function is referring to a wrong layer resource type
    LOG.debug("The used reference %s is not a Layer Version resource.", resolved_layer.value)
    raise InvalidResourceLinkingException(
        f"Could not use the value {resolved_layer.value} as a Layer for lambda function "
        f"{function_tf_resource.full_address}. Lambda Function Layer value should refer to valid lambda layer ARN "
        f"property"
    )


def _update_mapped_lambda_function_with_resolved_layers(
    cfn_functions: List[Dict],
    layers: List[Dict[str, str]],
    tf_layers: Dict[str, Dict],
) -> None:
    """
    Set The resolved layers list to the mapped lambda functions.

    Parameters
    ----------
    cfn_functions: List[Dict]
        A list of mapped lambda functions that are equivalent to the input terraform configuration lambda function

    layers: List
        The resolved layers values that will be used as a value for the mapped CFN function Layers attribute.

    tf_layers: Dict[str, Dict]
        Dictionary of all actual terraform layers resources (not configuration resources). The dictionary's key is the
        calculated logical id for each resource
    """
    LOG.debug("Set the resolved layers %s to the cfn functions %s", layers, cfn_functions)
    for cfn_func in cfn_functions:
        LOG.debug("Process the Lambda function %s", cfn_func)
        # Add the resolved layers list as it is to the mapped function that does not have any layers defined
        if not cfn_func["Properties"].get("Layers"):
            LOG.debug("The Lambda function %s does not have any layers defined.", cfn_func)
            cfn_func["Properties"]["Layers"] = layers
            continue

        # Check if the the mapped function layers list contains any arn value for one of the resolved layers to replace
        # it.
        for layer in layers:
            # resolve the layer arn string to check if it is already there in the CFN func layer property
            # layer logical id will be always in tf_layers, as we do not consider the references to layers that does not
            # exist in the tf_layers list as it means that this layer will not be created.
            LOG.debug("Check if the layer %s is already defined in function % layers.", layer, cfn_func)
            layer_arn = tf_layers[layer["Ref"]].get("values", {}).get("arn")

            # The resolved layer is a reference to a layers which is not applied yet, so there is no ARN value yet.
            if not layer_arn:
                LOG.debug("The layer %s is not applied yet, and does not have ARN property.", layer)
                cfn_func["Properties"]["Layers"].append(layer)
                continue

            # try to find a layer arn that equals the resolved layer arn so we can replace it with Ref value.
            try:
                layer_index = cfn_func["Properties"].get("Layers", []).index(layer_arn)
                LOG.debug(
                    "The layer %s has the arn value %s that exists in function %s layers.", layer, layer_arn, cfn_func
                )
                cfn_func["Properties"]["Layers"][layer_index] = layer
            except ValueError:
                # there is no matching layer ARN.
                LOG.debug(
                    "The layer %s has the arn value %s that does not exist in function %s layers.",
                    layer,
                    layer_arn,
                    cfn_func,
                )
                cfn_func["Properties"]["Layers"].append(layer)



"""
Constants related to the Terraform prepare hook.
"""
from samcli.lib.utils.resources import AWS_LAMBDA_FUNCTION as CFN_AWS_LAMBDA_FUNCTION
from samcli.lib.utils.resources import AWS_LAMBDA_LAYERVERSION as CFN_AWS_LAMBDA_LAYER_VERSION

SAM_METADATA_RESOURCE_NAME_ATTRIBUTE = "resource_name"

CFN_CODE_PROPERTIES = {
    CFN_AWS_LAMBDA_FUNCTION: "Code",
    CFN_AWS_LAMBDA_LAYER_VERSION: "Content",
}



"""
Utilities for code signing process
"""

import logging
from typing import Dict, List, Set

from click import STRING, prompt

from samcli.lib.providers.provider import Stack
from samcli.lib.providers.sam_function_provider import SamFunctionProvider

LOG = logging.getLogger(__name__)


def prompt_profile_name(profile_name, start_bold, end_bold):
    return prompt(f"\t{start_bold}Signing Profile Name{end_bold}", type=STRING, default=profile_name)


def prompt_profile_owner(profile_owner, start_bold, end_bold):
    # click requires to have non None value for passing
    if not profile_owner:
        profile_owner = ""

    profile_owner = prompt(
        f"\t{start_bold}Signing Profile Owner Account ID (optional){end_bold}",
        type=STRING,
        default=profile_owner,
        show_default=len(profile_owner) > 0,
    )

    return profile_owner


def extract_profile_name_and_owner_from_existing(function_or_layer_name, signing_profiles):
    profile_name = None
    profile_owner = None
    # extract any code sign config that is passed via command line
    if function_or_layer_name in signing_profiles:
        profile_name = signing_profiles[function_or_layer_name]["profile_name"]
        profile_owner = signing_profiles[function_or_layer_name]["profile_owner"]

    return profile_name, profile_owner


def signer_config_per_function(stacks: List[Stack]):
    functions_with_code_sign = set()
    layers_with_code_sign: Dict[str, Set[str]] = {}

    sam_functions = SamFunctionProvider(stacks)

    for sam_function in sam_functions.get_all():
        if sam_function.codesign_config_arn:
            function_name = sam_function.name
            LOG.debug("Found the following function with a code signing config %s", function_name)
            functions_with_code_sign.add(function_name)

            if sam_function.layers:
                for layer in sam_function.layers:
                    layer_name = layer.name
                    LOG.debug("Found following layers inside the function %s", layer_name)
                    if layer_name in layers_with_code_sign:
                        layers_with_code_sign[layer_name].add(function_name)
                    else:
                        functions_that_is_referring_to_function = set()
                        functions_that_is_referring_to_function.add(function_name)
                        layers_with_code_sign[layer_name] = functions_that_is_referring_to_function

    return functions_with_code_sign, layers_with_code_sign



"""
Delete a SAM stack
"""
import json
import logging
from typing import Optional

import boto3
import click
from click import confirm, prompt

from samcli.cli.cli_config_file import TomlProvider
from samcli.cli.context import Context
from samcli.commands.delete.exceptions import CfDeleteFailedStatusError
from samcli.lib.bootstrap.companion_stack.companion_stack_builder import CompanionStack
from samcli.lib.delete.cfn_utils import CfnUtils
from samcli.lib.package.artifact_exporter import Template
from samcli.lib.package.ecr_uploader import ECRUploader
from samcli.lib.package.local_files_utils import get_uploaded_s3_object_name
from samcli.lib.package.s3_uploader import S3Uploader
from samcli.lib.package.uploaders import Uploaders
from samcli.lib.utils.boto_utils import get_boto_config_with_user_agent

CONFIG_COMMAND = "deploy"
CONFIG_SECTION = "parameters"
TEMPLATE_STAGE = "Original"

LOG = logging.getLogger(__name__)


class DeleteContext:
    # TODO: Separate this context into 2 separate contexts guided and non-guided, just like deploy.
    def __init__(
        self,
        stack_name: str,
        region: str,
        profile: str,
        config_file: str,
        config_env: str,
        no_prompts: bool,
        s3_bucket: Optional[str],
        s3_prefix: Optional[str],
    ):
        self.stack_name = stack_name
        self.region = region
        self.profile = profile
        self.config_file = config_file
        self.config_env = config_env
        self.no_prompts = no_prompts
        self.s3_bucket = s3_bucket
        self.s3_prefix = s3_prefix
        self.cf_utils = None
        self.s3_uploader = None
        self.ecr_uploader = None
        self.uploaders = None
        self.cf_template_file_name = None
        self.delete_artifacts_folder = None
        self.delete_cf_template_file = None
        self.companion_stack_name = None

    def __enter__(self):
        self.parse_config_file()
        if not self.stack_name:
            LOG.debug("No stack-name input found")
            if not self.no_prompts:
                self.stack_name = prompt(
                    click.style("\tEnter stack name you want to delete", bold=True), type=click.STRING
                )
            else:
                raise click.BadOptionUsage(
                    option_name="--stack-name",
                    message="Missing option '--stack-name', provide a stack name that needs to be deleted.",
                )

        self.init_clients()
        return self

    def __exit__(self, *args):
        pass

    def parse_config_file(self):
        """
        Read the provided config file if it exists and assign the options values.
        """
        toml_provider = TomlProvider(CONFIG_SECTION, [CONFIG_COMMAND])
        config_options = toml_provider(
            config_path=self.config_file, config_env=self.config_env, cmd_names=[CONFIG_COMMAND]
        )
        if not config_options:
            return

        if not self.stack_name:
            self.stack_name = config_options.get("stack_name", None)
        # If the stack_name is same as the one present in samconfig file,
        # get the information about parameters if not specified by user.
        if self.stack_name and self.stack_name == config_options.get("stack_name", None):
            LOG.debug("Local config present and using the defined options")
            if not self.region:
                self.region = config_options.get("region", None)
            if not self.profile:
                self.profile = config_options.get("profile", None)
            if not self.s3_bucket:
                self.s3_bucket = config_options.get("s3_bucket", None)
            if not self.s3_prefix:
                self.s3_prefix = config_options.get("s3_prefix", None)

    def init_clients(self):
        """
        Initialize all the clients being used by sam delete.
        """
        if not self.region:
            if not self.no_prompts:
                session = boto3.Session()
                region = session.region_name
                self.region = region if region else "us-east-1"
            else:
                # TODO: as part of the guided and non-guided context separation, we need also to move the options
                # validations to a validator similar to samcli/lib/cli_validation/image_repository_validation.py.
                raise click.BadOptionUsage(
                    option_name="--region",
                    message="Missing option '--region', region is required to run the non guided delete command.",
                )

        if self.profile:
            Context.get_current_context().profile = self.profile
        if self.region:
            Context.get_current_context().region = self.region

        boto_config = get_boto_config_with_user_agent()

        # Define cf_client based on the region as different regions can have same stack-names
        cloudformation_client = boto3.client(
            "cloudformation", region_name=self.region if self.region else None, config=boto_config
        )

        s3_client = boto3.client("s3", region_name=self.region if self.region else None, config=boto_config)
        ecr_client = boto3.client("ecr", region_name=self.region if self.region else None, config=boto_config)

        self.s3_uploader = S3Uploader(s3_client=s3_client, bucket_name=self.s3_bucket, prefix=self.s3_prefix)

        self.ecr_uploader = ECRUploader(docker_client=None, ecr_client=ecr_client, ecr_repo=None, ecr_repo_multi=None)

        self.uploaders = Uploaders(self.s3_uploader, self.ecr_uploader)
        self.cf_utils = CfnUtils(cloudformation_client)

    def s3_prompts(self):
        """
        Guided prompts asking user to delete s3 artifacts
        """
        # Note: s3_bucket and s3_prefix information is only
        # available if it is provided as an option flag, a
        # local toml file or if this information is obtained
        # from the template resources and so if this
        # information is not found, warn the user that S3 artifacts
        # will need to be manually deleted.

        if not self.no_prompts and self.s3_bucket:
            if self.s3_prefix:
                self.delete_artifacts_folder = confirm(
                    click.style(
                        "\tAre you sure you want to delete the folder"
                        f" {self.s3_prefix} in S3 which contains the artifacts?",
                        bold=True,
                    ),
                    default=False,
                )
            if not self.delete_artifacts_folder:
                LOG.debug("S3 prefix not present or user does not want to delete the prefix folder")
                self.delete_cf_template_file = confirm(
                    click.style(
                        "\tDo you want to delete the template file" + f" {self.cf_template_file_name} in S3?", bold=True
                    ),
                    default=False,
                )
        elif self.s3_bucket:
            if self.s3_prefix:
                self.delete_artifacts_folder = True
            else:
                self.delete_cf_template_file = True

    def ecr_companion_stack_prompts(self):
        """
        User prompt to delete the ECR companion stack.
        """
        click.echo(f"\tFound ECR Companion Stack {self.companion_stack_name}")
        if self.no_prompts:
            return True

        return confirm(
            click.style(
                "\tDo you you want to delete the ECR companion stack"
                f" {self.companion_stack_name} in the region {self.region} ?",
                bold=True,
            ),
            default=False,
        )

    def ecr_repos_prompts(self, template: Template):
        """
        User prompts to delete the ECR repositories for the given template.

        :param template: Template to get the ECR repositories.
        """
        retain_repos = []
        ecr_repos = template.get_ecr_repos()

        if not self.no_prompts:
            for logical_id in ecr_repos:
                # Get all the repos from the companion stack
                repo = ecr_repos[logical_id]
                repo_name = repo["Repository"]

                delete_repo = confirm(
                    click.style(
                        f"\tECR repository {repo_name}"
                        " may not be empty. Do you want to delete the repository and all the images in it ?",
                        bold=True,
                    ),
                    default=False,
                )
                if not delete_repo:
                    retain_repos.append(logical_id)
        return retain_repos

    def delete_ecr_companion_stack(self):
        """
        Delete the ECR companion stack and ECR repositories based
        on user input.
        """
        delete_ecr_companion_stack_prompt = self.ecr_companion_stack_prompts()
        if delete_ecr_companion_stack_prompt or self.no_prompts:
            cf_ecr_companion_stack = self.cf_utils.get_stack_template(self.companion_stack_name, TEMPLATE_STAGE)
            ecr_stack_template_str = cf_ecr_companion_stack.get("TemplateBody", None)
            ecr_stack_template_str = json.dumps(ecr_stack_template_str, indent=4, ensure_ascii=False)

            ecr_companion_stack_template = Template(
                template_path=None,
                parent_dir=None,
                uploaders=self.uploaders,
                code_signer=None,
                template_str=ecr_stack_template_str,
            )

            retain_repos = self.ecr_repos_prompts(ecr_companion_stack_template)
            # Delete the repos created by ECR companion stack if not retained
            ecr_companion_stack_template.delete(retain_resources=retain_repos)

            click.echo(f"\t- Deleting ECR Companion Stack {self.companion_stack_name}")
            try:
                # If delete_stack fails and its status changes to DELETE_FAILED, retain
                # the user input repositories and delete the stack.
                self.cf_utils.delete_stack(stack_name=self.companion_stack_name)
                self.cf_utils.wait_for_delete(stack_name=self.companion_stack_name)
                LOG.debug("Deleted ECR Companion Stack: %s", self.companion_stack_name)

            except CfDeleteFailedStatusError:
                LOG.debug("delete_stack resulted failed and so re-try with retain_resources")
                self.cf_utils.delete_stack(stack_name=self.companion_stack_name, retain_resources=retain_repos)
                self.cf_utils.wait_for_delete(stack_name=self.companion_stack_name)

    def delete(self):
        """
        Delete method calls for Cloudformation stacks and S3 and ECR artifacts
        """
        # Fetch the template using the stack-name
        cf_template = self.cf_utils.get_stack_template(self.stack_name, TEMPLATE_STAGE)
        template_str = cf_template.get("TemplateBody", None)

        if isinstance(template_str, dict):
            template_str = json.dumps(template_str, indent=4, ensure_ascii=False)

        # Get the cloudformation template name using template_str
        self.cf_template_file_name = get_uploaded_s3_object_name(file_content=template_str, extension="template")

        template = Template(
            template_path=None,
            parent_dir=None,
            uploaders=self.uploaders,
            code_signer=None,
            template_str=template_str,
        )

        # If s3 info is not available, try to obtain it from CF
        # template resources.
        if not self.s3_bucket:
            s3_info = template.get_s3_info()
            self.s3_bucket = s3_info["s3_bucket"]
            self.s3_uploader.bucket_name = self.s3_bucket

            self.s3_prefix = s3_info["s3_prefix"]
            self.s3_uploader.prefix = self.s3_prefix

        self.s3_prompts()

        retain_resources = self.ecr_repos_prompts(template)

        # ECR companion stack delete prompts, if it exists
        companion_stack = CompanionStack(self.stack_name)

        ecr_companion_stack_exists = self.cf_utils.has_stack(stack_name=companion_stack.stack_name)
        if ecr_companion_stack_exists:
            LOG.debug("ECR Companion stack found for the input stack")
            self.companion_stack_name = companion_stack.stack_name
            self.delete_ecr_companion_stack()

        # Delete the artifacts and retain resources user selected not to delete
        template.delete(retain_resources=retain_resources)

        # Delete the CF template file in S3
        if self.delete_cf_template_file:
            self.s3_uploader.delete_artifact(remote_path=self.cf_template_file_name)

        # Delete the folder of artifacts if s3_bucket and s3_prefix provided
        elif self.delete_artifacts_folder:
            self.s3_uploader.delete_prefix_artifacts()

        # Delete the primary input stack
        try:
            click.echo(f"\t- Deleting Cloudformation stack {self.stack_name}")
            self.cf_utils.delete_stack(stack_name=self.stack_name)
            self.cf_utils.wait_for_delete(self.stack_name)
            LOG.debug("Deleted Cloudformation stack: %s", self.stack_name)

        except CfDeleteFailedStatusError:
            LOG.debug("delete_stack resulted failed and so re-try with retain_resources")
            self.cf_utils.delete_stack(stack_name=self.stack_name, retain_resources=retain_resources)
            self.cf_utils.wait_for_delete(self.stack_name)

        # Warn the user that s3 information is missing and to use --s3 options
        if not self.s3_bucket:
            LOG.debug("Cannot delete s3 objects as bucket is missing")
            click.secho(
                "\nWarning: Cannot resolve s3 bucket information from command options"
                " , local config file or cloudformation template. Please use"
                " --s3-bucket next time and"
                " delete s3 files manually if required.",
                fg="yellow",
            )

    def run(self):
        """
        Delete the stack based on the argument provided by user and samconfig.toml.
        """
        if not self.no_prompts:
            delete_stack = confirm(
                click.style(
                    f"\tAre you sure you want to delete the stack {self.stack_name}" f" in the region {self.region} ?",
                    bold=True,
                ),
                default=False,
            )

        if self.no_prompts or delete_stack:
            is_deployed = self.cf_utils.has_stack(stack_name=self.stack_name)
            # Check if the provided stack-name exists
            if is_deployed:
                LOG.debug("Input stack is deployed, continue deleting")
                self.delete()
                click.echo("\nDeleted successfully")
            else:
                LOG.debug("Input stack does not exists on Cloudformation")
                click.echo(
                    f"Error: The input stack {self.stack_name} does"
                    f" not exist on Cloudformation in the region {self.region}"
                )



"""
Display the Outputs of a SAM stack
"""
import logging
from typing import Optional

from samcli.commands.list.cli_common.list_common_context import ListContext
from samcli.lib.list.list_interfaces import ProducersEnum
from samcli.lib.list.mapper_consumer_factory import MapperConsumerFactory
from samcli.lib.list.stack_outputs.stack_outputs_producer import StackOutputsProducer

LOG = logging.getLogger(__name__)


class StackOutputsContext(ListContext):
    def __init__(self, stack_name: str, output: str, region: Optional[str], profile: Optional[str]):
        super().__init__()
        self.stack_name = stack_name
        self.output = output
        self.region = region
        self.profile = profile
        self.cloudformation_client = None

    def __enter__(self):
        self.init_clients()
        return self

    def __exit__(self, *args):
        pass

    def run(self) -> None:
        """
        Get the stack outputs for a stack
        """
        factory = MapperConsumerFactory()
        container = factory.create(producer=ProducersEnum.STACK_OUTPUTS_PRODUCER, output=self.output)

        producer = StackOutputsProducer(
            stack_name=self.stack_name,
            output=self.output,
            region=self.region,
            cloudformation_client=self.cloudformation_client,
            mapper=container.mapper,
            consumer=container.consumer,
        )
        producer.produce()



"""
This module contains utility functions for boto3 library
"""
from typing import Any, Optional

from boto3 import Session
from botocore.config import Config
from botocore.exceptions import ClientError
from typing_extensions import Protocol

from samcli import __version__
from samcli.cli.global_config import GlobalConfig


def get_boto_config_with_user_agent(**kwargs) -> Config:
    """
    Automatically add user agent string to boto configs.

    Parameters
    ----------
    kwargs :
        key=value params which will be added to the Config object

    Returns
    -------
    Config
        Returns config instance which contains given parameters in it
    """
    gc = GlobalConfig()
    return Config(
        user_agent_extra=f"aws-sam-cli/{__version__}/{gc.installation_id}"
        if gc.telemetry_enabled
        else f"aws-sam-cli/{__version__}",
        **kwargs,
    )


# Type definition of following boto providers, which is equal to Callable[[str], Any]
class BotoProviderType(Protocol):
    def __call__(self, service_name: str) -> Any:
        ...  # pragma: no cover


def get_boto_client_provider_from_session_with_config(session: Session, **kwargs) -> BotoProviderType:
    """
    Returns a wrapper function for boto client with given configuration. It can be used like;

    client_provider = get_boto_client_wrapper_with_config(session=session)
    lambda_client = client_provider("lambda")

    Parameters
    ----------
    session: Session
        Boto3 session object
    kwargs :
        Key-value params that will be passed to get_boto_config_with_user_agent

    Returns
    -------
        A callable function which will return a boto client
    """
    return lambda client_name: session.client(client_name, config=get_boto_config_with_user_agent(**kwargs))


def get_boto_client_provider_with_config(
    region: Optional[str] = None, profile: Optional[str] = None, **kwargs
) -> BotoProviderType:
    """
    Returns a wrapper function for boto client with given configuration. It can be used like;

    client_provider = get_boto_client_wrapper_with_config(region_name=region)
    lambda_client = client_provider("lambda")

    Parameters
    ----------
    region: Optional[str]
        AWS region name
    profile: Optional[str]
        Profile name from credentials
    kwargs :
        Key-value params that will be passed to get_boto_config_with_user_agent

    Returns
    -------
        A callable function which will return a boto client
    """
    return get_boto_client_provider_from_session_with_config(
        Session(region_name=region, profile_name=profile), **kwargs
    )


def get_boto_resource_provider_from_session_with_config(session: Session, **kwargs) -> BotoProviderType:
    """
    Returns a wrapper function for boto resource with given configuration. It can be used like;

    resource_provider = get_boto_resource_wrapper_with_config(session=session)
    cloudformation_resource = resource_provider("cloudformation")

    Parameters
    ----------
    session: Session
        Boto3 session object
    kwargs :
        Key-value params that will be passed to get_boto_config_with_user_agent

    Returns
    -------
        A callable function which will return a boto resource
    """
    return lambda resource_name: session.resource(resource_name, config=get_boto_config_with_user_agent(**kwargs))


def get_boto_resource_provider_with_config(
    region: Optional[str] = None, profile: Optional[str] = None, **kwargs
) -> BotoProviderType:
    """
    Returns a wrapper function for boto resource with given configuration. It can be used like;

    resource_provider = get_boto_resource_wrapper_with_config(region_name=region)
    cloudformation_resource = resource_provider("cloudformation")

    Parameters
    ----------
    region: Optional[str]
        AWS region name
    profile: Optional[str]
        Profile name from credentials
    kwargs :
        Key-value params that will be passed to get_boto_config_with_user_agent

    Returns
    -------
        A callable function which will return a boto resource
    """
    return get_boto_resource_provider_from_session_with_config(
        Session(region_name=region, profile_name=profile), **kwargs
    )


def get_client_error_code(client_error: ClientError) -> Optional[str]:
    """Extracts error code from boto ClientError"""
    return client_error.response.get("Error", {}).get("Code")



"""
CloudWatch log event puller implementation
"""
import logging
import time
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from botocore.exceptions import ClientError

from samcli.lib.observability.cw_logs.cw_log_event import CWLogEvent
from samcli.lib.observability.observability_info_puller import ObservabilityEventConsumer, ObservabilityPuller
from samcli.lib.utils.time import to_datetime, to_timestamp

LOG = logging.getLogger(__name__)


class CWLogPuller(ObservabilityPuller):
    """
    Puller implementation that can pull events from CloudWatch log group
    """

    def __init__(
        self,
        logs_client: Any,
        consumer: ObservabilityEventConsumer,
        cw_log_group: str,
        resource_name: Optional[str] = None,
        max_retries: int = 1000,
        poll_interval: int = 1,
    ):
        """
        Parameters
        ----------
        logs_client: CloudWatchLogsClient
            boto3 logs client instance
        consumer : ObservabilityEventConsumer
            Consumer instance that will process pulled events
        cw_log_group : str
            CloudWatch log group name
        resource_name : Optional[str]
            Optional parameter to assign a resource name for each event.
        max_retries: int
            Optional parameter to set maximum retries when tailing. Default value is 1000
        poll_interval: int
            Optional parameter to define sleep interval between pulling new log events when tailing. Default value is 1
        """
        self.logs_client = logs_client
        self.consumer = consumer
        self.cw_log_group = cw_log_group
        self.resource_name = resource_name
        self._max_retries = max_retries
        self._poll_interval = poll_interval
        self.latest_event_time = 0
        self.had_data = False
        self._invalid_log_group = False

    def tail(self, start_time: Optional[datetime] = None, filter_pattern: Optional[str] = None):
        if start_time:
            self.latest_event_time = to_timestamp(start_time)

        counter = self._max_retries
        while counter > 0 and not self.cancelled:
            LOG.debug("Tailing logs from %s starting at %s", self.cw_log_group, str(self.latest_event_time))

            counter -= 1
            try:
                self.load_time_period(to_datetime(self.latest_event_time), filter_pattern=filter_pattern)
            except ClientError as err:
                error_code = err.response.get("Error", {}).get("Code")
                if error_code == "ThrottlingException":
                    # if throttled, increase poll interval by 1 second each time
                    if self._poll_interval == 1:
                        self._poll_interval += 1
                    else:
                        self._poll_interval **= 2
                    LOG.warning(
                        "Throttled by CloudWatch Logs API, consider pulling logs for certain resources. "
                        "Increasing the poll interval time for resource %s to %s seconds",
                        self.cw_log_group,
                        self._poll_interval,
                    )
                else:
                    # if error is other than throttling, re-raise it
                    LOG.error("Failed while fetching new log events", exc_info=err)
                    raise err

            # This poll fetched logs. Reset the retry counter and set the timestamp for next poll
            if self.had_data:
                counter = self._max_retries
                self.latest_event_time += 1  # one extra millisecond to fetch next log event
                self.had_data = False

            # We already fetched logs once. Sleep for some time before querying again.
            # This also helps us scoot under the TPS limit for CloudWatch API call.
            time.sleep(self._poll_interval)

    def load_time_period(
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        filter_pattern: Optional[str] = None,
    ):
        kwargs = {"logGroupName": self.cw_log_group, "interleaved": True}

        if start_time:
            kwargs["startTime"] = to_timestamp(start_time)

        if end_time:
            kwargs["endTime"] = to_timestamp(end_time)

        if filter_pattern:
            kwargs["filterPattern"] = filter_pattern

        while True:
            LOG.debug("Fetching logs from CloudWatch with parameters %s", kwargs)
            try:
                result = self.logs_client.filter_log_events(**kwargs)
                self._invalid_log_group = False
            except self.logs_client.exceptions.ResourceNotFoundException:
                if not self._invalid_log_group:
                    LOG.debug(
                        "The specified log group %s does not exist. "
                        "This may be due to your resource have not been invoked yet.",
                        self.cw_log_group,
                    )
                    self._invalid_log_group = True
                break

            # Several events will be returned. Consume one at a time
            for event in result.get("events", []):
                self.had_data = True
                cw_event = CWLogEvent(self.cw_log_group, dict(event), self.resource_name)

                if cw_event.timestamp > self.latest_event_time:
                    self.latest_event_time = cw_event.timestamp

                self.consumer.consume(cw_event)

            # Keep iterating until there are no more logs left to query.
            next_token = result.get("nextToken", None)
            kwargs["nextToken"] = next_token
            if not next_token:
                break

    def load_events(self, event_ids: Union[List[Any], Dict]):
        LOG.debug("Loading specific events are not supported via CloudWatch Log Group")



"""
CLI command for "local start-lambda" command
"""

import logging

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, pass_context, print_cmdline_args
from samcli.cli.main import common_options as cli_framework_options
from samcli.commands._utils.experimental import ExperimentalFlag, experimental, is_experimental_enabled
from samcli.commands._utils.option_value_processor import process_image_options
from samcli.commands._utils.options import (
    generate_next_command_recommendation,
    hook_name_click_option,
    skip_prepare_infra_option,
)
from samcli.commands.local.cli_common.options import (
    invoke_common_options,
    local_common_options,
    service_common_options,
    warm_containers_common_options,
)
from samcli.commands.local.lib.exceptions import InvalidIntermediateImageError
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version
from samcli.local.docker.exceptions import ContainerNotStartableException

LOG = logging.getLogger(__name__)

HELP_TEXT = """
You can use this command to programmatically invoke your Lambda function locally using the AWS CLI or SDKs.
This command starts a local endpoint that emulates the AWS Lambda service, and you can run your automated
tests against this local Lambda endpoint. When you send an invoke to this endpoint using the AWS CLI or
SDK, it will locally execute the Lambda function specified in the request.\n
\b
SETUP
------
Start the local Lambda endpoint by running this command in the directory that contains your AWS SAM template.
$ sam local start-lambda\n
\b
USING AWS CLI
-------------
Then, you can invoke your Lambda function locally using the AWS CLI
$ aws lambda invoke --function-name "HelloWorldFunction" --endpoint-url "http://127.0.0.1:3001" --no-verify-ssl out.txt
\n
\b
USING AWS SDK
-------------
You can also use the AWS SDK in your automated tests to invoke your functions programatically.
Here is a Python example:
    self.lambda_client = boto3.client('lambda',
                                  endpoint_url="http://127.0.0.1:3001",
                                  use_ssl=False,
                                  verify=False,
                                  config=Config(signature_version=UNSIGNED,
                                                read_timeout=0,
                                                retries={'max_attempts': 0}))
    self.lambda_client.invoke(FunctionName="HelloWorldFunction")
"""


@click.command(
    "start-lambda",
    help=HELP_TEXT,
    short_help="Starts a local endpoint you can use to invoke your local Lambda functions.",
)
@configuration_option(provider=TomlProvider(section="parameters"))
@hook_name_click_option(
    force_prepare=False, invalid_coexist_options=["t", "template-file", "template", "parameter-overrides"]
)
@skip_prepare_infra_option
@service_common_options(3001)
@invoke_common_options
@experimental
@warm_containers_common_options
@local_common_options
@cli_framework_options
@aws_creds_options
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
def cli(
    ctx,  # pylint: disable=R0914
    # start-lambda Specific Options
    host,
    port,
    # Common Options for Lambda Invoke
    template_file,
    env_vars,
    debug_port,
    debug_args,
    debugger_path,
    container_env_vars,
    docker_volume_basedir,
    docker_network,
    log_file,
    layer_cache_basedir,
    skip_pull_image,
    force_image_build,
    parameter_overrides,
    config_file,
    config_env,
    warm_containers,
    shutdown,
    debug_function,
    container_host,
    container_host_interface,
    invoke_image,
    hook_name,
    skip_prepare_infra,
):
    """
    `sam local start-lambda` command entry point
    """
    # All logic must be implemented in the ``do_cli`` method. This helps with easy unit testing

    do_cli(
        ctx,
        host,
        port,
        template_file,
        env_vars,
        debug_port,
        debug_args,
        debugger_path,
        container_env_vars,
        docker_volume_basedir,
        docker_network,
        log_file,
        layer_cache_basedir,
        skip_pull_image,
        force_image_build,
        parameter_overrides,
        warm_containers,
        shutdown,
        debug_function,
        container_host,
        container_host_interface,
        invoke_image,
        hook_name,
    )  # pragma: no cover


def do_cli(  # pylint: disable=R0914
    ctx,
    host,
    port,
    template,
    env_vars,
    debug_port,
    debug_args,
    debugger_path,
    container_env_vars,
    docker_volume_basedir,
    docker_network,
    log_file,
    layer_cache_basedir,
    skip_pull_image,
    force_image_build,
    parameter_overrides,
    warm_containers,
    shutdown,
    debug_function,
    container_host,
    container_host_interface,
    invoke_image,
    hook_name,
):
    """
    Implementation of the ``cli`` method, just separated out for unit testing purposes
    """

    from samcli.commands.local.cli_common.invoke_context import InvokeContext
    from samcli.commands.local.cli_common.user_exceptions import UserException
    from samcli.commands.local.lib.exceptions import OverridesNotWellDefinedError
    from samcli.commands.local.lib.local_lambda_service import LocalLambdaService
    from samcli.commands.validate.lib.exceptions import InvalidSamDocumentException
    from samcli.lib.providers.exceptions import InvalidLayerReference
    from samcli.local.docker.lambda_debug_settings import DebuggingNotSupported

    if (
        hook_name
        and ExperimentalFlag.IaCsSupport.get(hook_name) is not None
        and not is_experimental_enabled(ExperimentalFlag.IaCsSupport.get(hook_name))
    ):
        LOG.info("Terraform Support beta feature is not enabled.")
        return

    LOG.debug("local start_lambda command is called")

    processed_invoke_images = process_image_options(invoke_image)

    # Pass all inputs to setup necessary context to invoke function locally.
    # Handler exception raised by the processor for invalid args and print errors

    try:
        with InvokeContext(
            template_file=template,
            function_identifier=None,  # Don't scope to one particular function
            env_vars_file=env_vars,
            docker_volume_basedir=docker_volume_basedir,
            docker_network=docker_network,
            log_file=log_file,
            skip_pull_image=skip_pull_image,
            debug_ports=debug_port,
            debug_args=debug_args,
            debugger_path=debugger_path,
            container_env_vars_file=container_env_vars,
            parameter_overrides=parameter_overrides,
            layer_cache_basedir=layer_cache_basedir,
            force_image_build=force_image_build,
            aws_region=ctx.region,
            aws_profile=ctx.profile,
            warm_container_initialization_mode=warm_containers,
            debug_function=debug_function,
            shutdown=shutdown,
            container_host=container_host,
            container_host_interface=container_host_interface,
            invoke_images=processed_invoke_images,
        ) as invoke_context:

            service = LocalLambdaService(lambda_invoke_context=invoke_context, port=port, host=host)
            service.start()
            command_suggestions = generate_next_command_recommendation(
                [
                    ("Validate SAM template", "sam validate"),
                    ("Test Function in the Cloud", "sam sync --stack-name {{stack-name}} --watch"),
                    ("Deploy", "sam deploy --guided"),
                ]
            )
            click.secho(command_suggestions, fg="yellow")

    except (
        InvalidSamDocumentException,
        OverridesNotWellDefinedError,
        InvalidLayerReference,
        InvalidIntermediateImageError,
        DebuggingNotSupported,
    ) as ex:
        raise UserException(str(ex), wrapped_from=ex.__class__.__name__) from ex
    except ContainerNotStartableException as ex:
        raise UserException(str(ex), wrapped_from=ex.__class__.__name__) from ex



"""
Construct one CLI page based on the input provided and returns customer choice
"""

import click


def do_paginate_cli(pages, page_to_be_rendered, items_per_page, is_last_page, cli_display_message):
    """
    Responsible for displaying a generic CLI page with available user choices for pagination/seletion
    :param pages:
    :param page_to_be_rendered:
    :param items_per_page:
    :param is_last_page:
    :param cli_display_message:
    :return: User decision on displayed page
    """
    options = pages.get(page_to_be_rendered)
    choice_num = page_to_be_rendered * items_per_page + 1
    choices = []

    # Track possible user choices (which are unique across all pages)
    for option in options:
        msg = str(choice_num) + " - " + option
        click.echo("\t" + msg)
        choices.append(choice_num)
        choice_num = choice_num + 1

    # Single page is available no pagination
    if len(pages) == 1 and is_last_page:
        message = str.format(cli_display_message["single_page"])
    elif not page_to_be_rendered:
        # (multi) First page
        choices = choices + ["N", "n"]
        message = cli_display_message["first_page"]
    elif is_last_page and page_to_be_rendered == len(pages) - 1:
        # (multi) Last page
        choices = choices + ["P", "p"]
        message = cli_display_message["last_page"]
    else:
        # (multi) Middle page
        choices = choices + ["N", "n", "P", "p"]
        message = cli_display_message["middle_page"]

    final_choices = list(map(str, choices))
    choice = click.prompt(message, type=click.Choice(final_choices), show_choices=False)

    # Return page to be rendered based on the user selection
    if choice in ("N", "n"):
        return {"choice": None, "page_to_render": page_to_be_rendered + 1}
    if choice in ("P", "p"):
        return {"choice": None, "page_to_render": page_to_be_rendered - 1}

    # Recalculate page index based on the global choice ID, and the number of items per page
    # e.g. If user picks choice '11', and we know there are 10 items per page,
    # then we know the user is on page 2
    index = int(choice) % items_per_page
    if index:
        index = index - 1
    else:
        index = items_per_page - 1
    return {"choice": options[index], "page_to_render": None}



"""
Contains helper functions for validation and exception handling of "sam logs" command
"""
from functools import wraps
from typing import Any, Callable, Dict

import click
from botocore.exceptions import ClientError
from click import BadOptionUsage, Context

from samcli.commands.exceptions import InvalidStackNameException
from samcli.lib.utils.boto_utils import get_client_error_code


def stack_name_cw_log_group_validation(func):
    """
    Wrapper Validation function that will run last after the all cli parmaters have been loaded
    to check for conditions surrounding `--stack-name` and `--cw-log-group`. The
    reason they are done last instead of in callback functions, is because the options depend
    on each other, and this breaks cyclic dependencies.

    :param func: Click command function
    :return: Click command function after validation
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        ctx = click.get_current_context()
        stack_name = ctx.params.get("stack_name")
        cw_log_groups = ctx.params.get("cw_log_group")
        names = ctx.params.get("name")

        # if --name is provided --stack-name should be provided as well
        if names and not stack_name:
            raise BadOptionUsage(
                option_name="--stack-name",
                ctx=ctx,
                message="Missing option. Please provide '--stack-name' when using '--name' option",
            )

        # either --stack-name or --cw-log-group flags should be provided
        if not stack_name and not cw_log_groups:
            raise BadOptionUsage(
                option_name="--stack-name",
                ctx=ctx,
                message="Missing option. Please provide '--stack-name' or '--cw-log-group'",
            )

        return func(*args, **kwargs)

    return wrapped


def _handle_client_error(ex: ClientError) -> None:
    """
    Handles client error which was caused by ListStackResources event
    """
    operation_name = ex.operation_name
    client_error_code = get_client_error_code(ex)
    if client_error_code == "ValidationError" and operation_name == "ListStackResources":
        click_context: Context = click.get_current_context()
        stack_name_value = click_context.params.get("stack_name")
        raise InvalidStackNameException(
            f"Invalid --stack-name parameter. Stack with id '{stack_name_value}' does not exist"
        )


SAM_LOGS_ADDITIONAL_EXCEPTION_HANDLERS: Dict[Any, Callable] = {ClientError: _handle_client_error}



"""
Terraform translate to CFN implementation

This method contains the logic required to translate the `terraform show` JSON output into a Cloudformation template
"""
import hashlib
import logging
from typing import Any, Dict, List, Tuple, Union

from samcli.hook_packages.terraform.hooks.prepare.constants import (
    CFN_CODE_PROPERTIES,
    SAM_METADATA_RESOURCE_NAME_ATTRIBUTE,
)
from samcli.hook_packages.terraform.hooks.prepare.enrich import enrich_resources_and_generate_makefile
from samcli.hook_packages.terraform.hooks.prepare.property_builder import (
    REMOTE_DUMMY_VALUE,
    RESOURCE_TRANSLATOR_MAPPING,
    TF_AWS_LAMBDA_FUNCTION,
    TF_AWS_LAMBDA_LAYER_VERSION,
    PropertyBuilderMapping,
)
from samcli.hook_packages.terraform.hooks.prepare.resource_linking import (
    _build_module,
    _get_configuration_address,
    _link_lambda_function_to_layer,
    _resolve_resource_attribute,
)
from samcli.hook_packages.terraform.hooks.prepare.types import (
    ConstantValue,
    References,
    ResolvedReference,
    SamMetadataResource,
    TFModule,
    TFResource,
)
from samcli.hook_packages.terraform.lib.utils import (
    _calculate_configuration_attribute_value_hash,
    build_cfn_logical_id,
    get_sam_metadata_planned_resource_value_attribute,
)
from samcli.lib.hook.exceptions import PrepareHookException
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.resources import AWS_LAMBDA_FUNCTION as CFN_AWS_LAMBDA_FUNCTION

SAM_METADATA_RESOURCE_TYPE = "null_resource"
SAM_METADATA_NAME_PREFIX = "sam_metadata_"

AWS_PROVIDER_NAME = "registry.terraform.io/hashicorp/aws"
NULL_RESOURCE_PROVIDER_NAME = "registry.terraform.io/hashicorp/null"

LOG = logging.getLogger(__name__)


def translate_to_cfn(tf_json: dict, output_directory_path: str, terraform_application_dir: str) -> dict:
    """
    Translates the json output of a terraform show into CloudFormation

    Parameters
    ----------
    tf_json: dict
        A terraform show json output
    output_directory_path: str
        the string path to write the metadata file and makefile
    terraform_application_dir: str
        the terraform project root directory

    Returns
    -------
    dict
        The CloudFormation resulting from translating tf_json
    """
    # setup root_module and cfn dict
    root_module = tf_json.get("planned_values", {}).get("root_module")
    cfn_dict: dict = {"AWSTemplateFormatVersion": "2010-09-09", "Resources": {}}
    if not root_module:
        return cfn_dict

    LOG.debug("Mapping Lambda functions to their corresponding layers.")
    input_vars: Dict[str, Union[ConstantValue, References]] = {
        var_name: ConstantValue(value=var_value.get("value"))
        for var_name, var_value in tf_json.get("variables", {}).items()
    }
    root_tf_module = _build_module("", tf_json.get("configuration", {}).get("root_module"), input_vars, None)

    # to map s3 object sources to respective functions later
    # this dictionary will map between the hash value of the S3 Bucket attributes, and a tuple of the planned value
    # source code path, and the configuration value of the source code path.
    s3_hash_to_source: Dict[str, Tuple[str, List[Union[ConstantValue, ResolvedReference]]]] = {}

    # map code/imageuri to Lambda resources
    # the key is the hash value of lambda code/imageuri
    # the value is the list of pair of the resource logical id, and the lambda cfn resource dict
    lambda_resources_to_code_map: Dict[str, List[Tuple[Dict, str]]] = {}

    sam_metadata_resources: List[SamMetadataResource] = []

    lambda_layers_terraform_resources: Dict[str, Dict] = {}
    lambda_funcs_conf_cfn_resources: Dict[str, List] = {}
    lambda_config_funcs_conf_cfn_resources: Dict[str, TFResource] = {}

    # create and iterate over queue of modules to handle child modules
    module_queue = [(root_module, root_tf_module)]
    while module_queue:
        modules_pair = module_queue.pop(0)
        curr_module, curr_tf_module = modules_pair
        curr_module_address = curr_module.get("address")

        _add_child_modules_to_queue(curr_module, curr_tf_module, module_queue)

        # iterate over resources for current module
        resources = curr_module.get("resources", {})
        for resource in resources:
            resource_provider = resource.get("provider_name")
            resource_type = resource.get("type")
            resource_values = resource.get("values")
            resource_full_address = resource.get("address")
            resource_name = resource.get("name")
            resource_mode = resource.get("mode")

            resource_address = (
                f"data.{resource_type}.{resource_name}"
                if resource_mode == "data"
                else f"{resource_type}.{resource_name}"
            )
            config_resource_address = _get_configuration_address(resource_address)
            if config_resource_address not in curr_tf_module.resources:
                raise PrepareHookException(
                    f"There is no configuration resource for resource address {resource_full_address} and "
                    f"configuration address {config_resource_address}"
                )

            config_resource = curr_tf_module.resources[config_resource_address]

            if (
                resource_provider == NULL_RESOURCE_PROVIDER_NAME
                and resource_type == SAM_METADATA_RESOURCE_TYPE
                and resource_name.startswith(SAM_METADATA_NAME_PREFIX)
            ):
                _add_metadata_resource_to_metadata_list(
                    SamMetadataResource(curr_module_address, resource, config_resource),
                    resource,
                    sam_metadata_resources,
                )
                continue

            # only process supported provider
            if resource_provider != AWS_PROVIDER_NAME:
                continue

            # store S3 sources
            if resource_type == "aws_s3_object":
                s3_bucket = (
                    resource_values.get("bucket")
                    if "bucket" in resource_values
                    else _resolve_resource_attribute(config_resource, "bucket")
                )
                s3_key = (
                    resource_values.get("key")
                    if "key" in resource_values
                    else _resolve_resource_attribute(config_resource, "key")
                )
                obj_hash = _get_s3_object_hash(s3_bucket, s3_key)
                code_artifact = resource_values.get("source")
                config_code_artifact = (
                    code_artifact if code_artifact else _resolve_resource_attribute(config_resource, "source")
                )
                s3_hash_to_source[obj_hash] = (code_artifact, config_code_artifact)

            resource_translator = RESOURCE_TRANSLATOR_MAPPING.get(resource_type)
            # resource type not supported
            if not resource_translator:
                continue

            # translate TF resource "values" to CFN properties
            LOG.debug("Processing resource %s", resource_full_address)
            translated_properties = _translate_properties(
                resource_values, resource_translator.property_builder_mapping, config_resource
            )
            translated_resource = {
                "Type": resource_translator.cfn_name,
                "Properties": translated_properties,
                "Metadata": {"SamResourceId": resource_full_address, "SkipBuild": True},
            }

            # build CFN logical ID from resource address
            logical_id = build_cfn_logical_id(resource_full_address)

            # Add resource to cfn dict
            cfn_dict["Resources"][logical_id] = translated_resource

            if resource_type == TF_AWS_LAMBDA_LAYER_VERSION:
                lambda_layers_terraform_resources[logical_id] = resource
                planned_value_layer_code_path = translated_properties.get("Content")
                _add_lambda_resource_code_path_to_code_map(
                    config_resource,
                    "layer",
                    lambda_resources_to_code_map,
                    logical_id,
                    planned_value_layer_code_path,
                    "filename",
                    translated_resource,
                )

            if resource_type == TF_AWS_LAMBDA_FUNCTION:
                resolved_config_address = _get_configuration_address(resource_full_address)
                matched_lambdas = lambda_funcs_conf_cfn_resources.get(resolved_config_address, [])
                matched_lambdas.append(translated_resource)
                lambda_funcs_conf_cfn_resources[resolved_config_address] = matched_lambdas
                lambda_config_funcs_conf_cfn_resources[resolved_config_address] = config_resource

                resource_type = translated_properties.get("PackageType", ZIP)
                resource_type_constants = {ZIP: ("zip", "filename"), IMAGE: ("image", "image_uri")}
                planned_value_function_code_path = (
                    translated_properties.get("Code")
                    if resource_type == ZIP
                    else translated_properties.get("Code", {}).get("ImageUri")
                )
                func_type, tf_code_property = resource_type_constants[resource_type]

                _add_lambda_resource_code_path_to_code_map(
                    config_resource,
                    func_type,
                    lambda_resources_to_code_map,
                    logical_id,
                    planned_value_function_code_path,
                    tf_code_property,
                    translated_resource,
                )

    # map s3 object sources to corresponding functions
    LOG.debug("Mapping S3 object sources to corresponding functions")
    _map_s3_sources_to_functions(s3_hash_to_source, cfn_dict.get("Resources", {}), lambda_resources_to_code_map)

    _link_lambda_functions_to_layers(
        lambda_config_funcs_conf_cfn_resources, lambda_funcs_conf_cfn_resources, lambda_layers_terraform_resources
    )

    if sam_metadata_resources:
        LOG.debug("Enrich the mapped resources with the sam metadata information and generate Makefile")
        enrich_resources_and_generate_makefile(
            sam_metadata_resources,
            cfn_dict.get("Resources", {}),
            output_directory_path,
            terraform_application_dir,
            lambda_resources_to_code_map,
        )
    else:
        LOG.debug("There is no sam metadata resources, no enrichment or Makefile is required")

    # check if there is still any dummy remote values for lambda resource imagesUri or S3 attributes
    _check_dummy_remote_values(cfn_dict.get("Resources", {}))

    return cfn_dict


def _add_child_modules_to_queue(curr_module: Dict, curr_module_configuration: TFModule, modules_queue: List) -> None:
    """
    Iterate over the children modules of current module and add each module with its related child module configuration
    to the modules_queue.

    Parameters
    ----------
    curr_module: Dict
        The current module in the planned values
    curr_module_configuration: TFModule
        The current module configuration
    modules_queue: List
        The list of modules
    """
    child_modules = curr_module.get("child_modules")
    if child_modules:
        for child_module in child_modules:
            config_child_module_address = (
                _get_configuration_address(child_module["address"]) if "address" in child_module else None
            )
            module_name = (
                config_child_module_address[config_child_module_address.rfind(".") + 1 :]
                if config_child_module_address
                else None
            )
            child_tf_module = curr_module_configuration.child_modules.get(module_name) if module_name else None
            if child_tf_module is None:
                raise PrepareHookException(
                    f"Module {config_child_module_address} exists in terraform planned_value, but does not exist "
                    "in terraform configuration"
                )
            modules_queue.append((child_module, child_tf_module))


def _add_metadata_resource_to_metadata_list(
    sam_metadata_resource: SamMetadataResource,
    sam_metadata_resource_planned_values: Dict,
    sam_metadata_resources: List[SamMetadataResource],
) -> None:
    """
    Prioritize the metadata resources that has resource name value to overwrite the metadata resources that does not
    have resource name value.

    Parameters
    ----------
    sam_metadata_resource: SamMetadataResource
        The mapped metadata resource
    sam_metadata_resource_planned_values: Dict
        The metadata resource in planned values section
    sam_metadata_resources: List[SamMetadataResource]
        The list of metadata resources
    """
    if get_sam_metadata_planned_resource_value_attribute(
        sam_metadata_resource_planned_values, SAM_METADATA_RESOURCE_NAME_ATTRIBUTE
    ):
        sam_metadata_resources.append(sam_metadata_resource)
    else:
        sam_metadata_resources.insert(0, sam_metadata_resource)


def _translate_properties(
    tf_properties: dict, property_builder_mapping: PropertyBuilderMapping, resource: TFResource
) -> dict:
    """
    Translates the properties of a terraform resource into the equivalent properties of a CloudFormation resource

    Parameters
    ----------
    tf_properties: dict
        The terraform properties to translate
    property_builder_mappping: PropertyBuilderMapping
        A mapping of the CloudFormation property name to a function for building that property
    resource: TFResource
        The terraform configuration resource that can be used to retrieve some attributes values if needed

    Returns
    -------
    dict
        The CloudFormation properties resulting from translating tf_properties
    """
    cfn_properties = {}
    for cfn_property_name, cfn_property_builder in property_builder_mapping.items():
        cfn_property_value = cfn_property_builder(tf_properties, resource)
        if cfn_property_value is not None:
            cfn_properties[cfn_property_name] = cfn_property_value
    return cfn_properties


def _add_lambda_resource_code_path_to_code_map(
    terraform_resource: TFResource,
    lambda_resource_prefix: str,
    lambda_resources_to_code_map: Dict,
    logical_id: str,
    lambda_resource_code_value: Any,
    terraform_code_property_name: str,
    translated_resource: Dict,
) -> None:
    """
    Calculate the hash value of  the lambda resource code path planned value or the configuration value and use it to
    map the lambda resource logical id to the source code path. This will be used later to map the metadata resource to
    the correct lambda resource.

    Parameters
    ----------
    terraform_resource: TFResource
        The mapped TF resource. This will be used to resolve the configuration value of the code attribute in the lambda
         resource
    lambda_resource_prefix: str
        a string prefix to be added to the hash value to differentiate between the different lambda resources types
    lambda_resources_to_code_map: dict
        the map between lambda resources code path values, and the lambda resources logical ids
    logical_id: str
        lambda resource logical id
    lambda_resource_code_value: Any
        The planned value of the lambda resource code path
    terraform_code_property_name: str
        The lambda resource code property name
    translated_resource: Dict
        The CFN translated lambda resource
    """
    if not lambda_resource_code_value or not isinstance(lambda_resource_code_value, str):
        lambda_resource_code_value = _resolve_resource_attribute(terraform_resource, terraform_code_property_name)
    if lambda_resource_code_value:
        hash_value = (
            f"{lambda_resource_prefix}_{_calculate_configuration_attribute_value_hash(lambda_resource_code_value)}"
        )
        functions_list = lambda_resources_to_code_map.get(hash_value, [])
        functions_list.append((translated_resource, logical_id))
        lambda_resources_to_code_map[hash_value] = functions_list


def _link_lambda_functions_to_layers(
    lambda_config_funcs_conf_cfn_resources: Dict[str, TFResource],
    lambda_funcs_conf_cfn_resources: Dict[str, List],
    lambda_layers_terraform_resources: Dict[str, Dict],
):
    """
    Iterate through all of the resources and link the corresponding Lambda Layers to each Lambda Function

    Parameters
    ----------
    lambda_config_funcs_conf_cfn_resources: Dict[str, TFResource]
        Dictionary of configuration lambda resources
    lambda_funcs_conf_cfn_resources: Dict[str, List]
        Dictionary containing resolved configuration addresses matched up to the cfn Lambda functions
    lambda_layers_terraform_resources: Dict[str, Dict]
        Dictionary of all actual terraform layers resources (not configuration resources). The dictionary's key is the
        calculated logical id for each resource

    Returns
    -------
    dict
        The CloudFormation resulting from translating tf_json
    """
    for config_address, resource in lambda_config_funcs_conf_cfn_resources.items():
        if config_address in lambda_funcs_conf_cfn_resources:
            LOG.debug("Linking layers for Lambda function %s", resource.full_address)
            _link_lambda_function_to_layer(
                resource, lambda_funcs_conf_cfn_resources[config_address], lambda_layers_terraform_resources
            )


def _map_s3_sources_to_functions(
    s3_hash_to_source: Dict[str, Tuple[str, List[Union[ConstantValue, ResolvedReference]]]],
    cfn_resources: Dict[str, Any],
    lambda_resources_to_code_map: Dict[str, List[Tuple[Dict, str]]],
) -> None:
    """
    Maps the source property of terraform AWS S3 object resources into the the Code property of
    CloudFormation AWS Lambda Function resources, and append the hash value of the artifacts path to the lambda
    resources code map.

    Parameters
    ----------
    s3_hash_to_source: Dict[str, Tuple[str, List[Union[ConstantValue, ResolvedReference]]]]
        Mapping of S3 object hash to S3 object source and the S3 Object configuration source value
    cfn_resources: dict
        CloudFormation resources
    lambda_resources_to_code_map: Dict
        the map between lambda resources code path values, and the lambda resources logical ids
    """
    for resource_logical_id, resource in cfn_resources.items():
        resource_type = resource.get("Type")
        if resource_type in CFN_CODE_PROPERTIES:
            code_property = CFN_CODE_PROPERTIES[resource_type]

            code = resource.get("Properties").get(code_property)

            # mapping not possible if function doesn't have bucket and key
            if isinstance(code, str):
                continue

            bucket = code.get("S3Bucket_config_value") if "S3Bucket_config_value" in code else code.get("S3Bucket")
            key = code.get("S3Key_config_value") if "S3Key_config_value" in code else code.get("S3Key")

            if bucket and key:
                obj_hash = _get_s3_object_hash(bucket, key)
                source = s3_hash_to_source.get(obj_hash)
                if source:
                    if source[0]:
                        tf_address = resource.get("Metadata", {}).get("SamResourceId")
                        LOG.debug(
                            "Found S3 object resource with matching bucket and key for function %s."
                            " Setting function's Code property to the matching S3 object's source: %s",
                            tf_address,
                            source[0],
                        )
                        resource["Properties"][code_property] = source[0]

                    references = source[0] or source[1]
                    res_type = "zip" if resource_type == CFN_AWS_LAMBDA_FUNCTION else "layer"
                    if references:
                        hash_value = f"{res_type}_{_calculate_configuration_attribute_value_hash(references)}"
                        resources_list = lambda_resources_to_code_map.get(hash_value, [])
                        resources_list.append((resource, resource_logical_id))
                        lambda_resources_to_code_map[hash_value] = resources_list


def _check_dummy_remote_values(cfn_resources: Dict[str, Any]) -> None:
    """
    Check if there is any lambda function/layer that has a dummy remote value for its code.imageuri or
    code.s3 attributes, and raise a validation error for it.

    Parameters
    ----------
    cfn_resources: dict
        CloudFormation resources
    """
    for _, resource in cfn_resources.items():
        resource_type = resource.get("Type")
        if resource_type in CFN_CODE_PROPERTIES:
            code_property = CFN_CODE_PROPERTIES[resource_type]

            code = resource.get("Properties").get(code_property)

            # there is no code property, this is the expected behaviour in image package type functions
            if code is None:
                continue

            # its value is a path to a local source code
            if isinstance(code, str):
                continue

            bucket = code.get("S3Bucket")
            key = code.get("S3Key")
            image_uri = code.get("ImageUri")

            if (bucket and bucket == REMOTE_DUMMY_VALUE) or (key and key == REMOTE_DUMMY_VALUE):
                raise PrepareHookException(
                    f"Lambda resource {resource.get('Metadata', {}).get('SamResourceId')} is referring to an S3 bucket "
                    f"that is not created yet, and there is no sam metadata resource set for it to build its code "
                    f"locally"
                )

            if image_uri and image_uri == REMOTE_DUMMY_VALUE:
                raise PrepareHookException(
                    f"Lambda resource {resource.get('Metadata', {}).get('SamResourceId')} is referring to an image uri "
                    "that is not created yet, and there is no sam metadata resource set for it to build its image "
                    "locally."
                )


def _get_s3_object_hash(
    bucket: Union[str, List[Union[ConstantValue, ResolvedReference]]],
    key: Union[str, List[Union[ConstantValue, ResolvedReference]]],
) -> str:
    """
    Creates a hash for an AWS S3 object out of the bucket and key

    Parameters
    ----------
    bucket: Union[str, List[Union[ConstantValue, ResolvedReference]]]
        bucket for the S3 object
    key: Union[str, List[Union[ConstantValue, ResolvedReference]]]
        key for the S3 object

    Returns
    -------
    str
        hash for the given bucket and key
    """
    md5 = hashlib.md5()
    md5.update(_calculate_configuration_attribute_value_hash(bucket).encode())
    md5.update(_calculate_configuration_attribute_value_hash(key).encode())
    # TODO: Hash version if it exists in addition to key and bucket
    return md5.hexdigest()



"""
Sets up the cli for resources
"""

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, common_options, pass_context, print_cmdline_args
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands._utils.options import template_option_without_build
from samcli.commands.list.cli_common.options import output_option, stack_name_not_provided_message, stack_name_option
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version

HELP_TEXT = """
Get a summary of the cloud endpoints in the stack.\n
This command will show both the cloud and local endpoints that can
be used with sam local and sam sync. Currently the endpoint resources
are Lambda functions and API Gateway API resources.
"""


@click.command(name="endpoints", help=HELP_TEXT)
@configuration_option(provider=TomlProvider(section="parameters"))
@stack_name_option
@output_option
@template_option_without_build
@aws_creds_options
@common_options
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
@command_exception_handler
def cli(self, stack_name, output, template_file, config_file, config_env):
    """
    `sam list endpoints` command entry point
    """
    do_cli(stack_name=stack_name, output=output, region=self.region, profile=self.profile, template_file=template_file)


def do_cli(stack_name, output, region, profile, template_file):
    """
    Implementation of the ``cli`` method
    """
    from samcli.commands.list.endpoints.endpoints_context import EndpointsContext

    with EndpointsContext(
        stack_name=stack_name, output=output, region=region, profile=profile, template_file=template_file
    ) as endpoints_context:
        if not stack_name:
            stack_name_not_provided_message()
        endpoints_context.run()



"""
Represents a manifest that lists the available SAM pipeline templates.
Example:
    providers:
      - displayName:Jenkins
        id: jenkins
      - displayName:Gitlab CI/CD
        id: gitlab
      - displayName:Github Actions
        id: github-actions
    templates:
      - displayName: jenkins-two-environments-pipeline
        provider: Jenkins
        location: templates/cookiecutter-jenkins-two-environments-pipeline
      - displayName: gitlab-two-environments-pipeline
        provider: Gitlab
        location: templates/cookiecutter-gitlab-two-environments-pipeline
      - displayName: Github-Actions-two-environments-pipeline
        provider: Github Actions
        location: templates/cookiecutter-github-actions-two-environments-pipeline
"""
from pathlib import Path
from typing import Dict, List

import yaml

from samcli.commands.exceptions import AppPipelineTemplateManifestException
from samcli.yamlhelper import parse_yaml_file


class Provider:
    """CI/CD system such as Jenkins, Gitlab and GitHub-Actions"""

    def __init__(self, manifest: Dict) -> None:
        self.id: str = manifest["id"]
        self.display_name: str = manifest["displayName"]


class PipelineTemplateMetadata:
    """The metadata of a Given pipeline template"""

    def __init__(self, manifest: Dict) -> None:
        self.display_name: str = manifest["displayName"]
        self.provider: str = manifest["provider"]
        self.location: str = manifest["location"]


class PipelineTemplatesManifest:
    """The metadata of the available CI/CD systems and the pipeline templates"""

    def __init__(self, manifest_path: Path) -> None:
        try:
            manifest: Dict = parse_yaml_file(file_path=str(manifest_path))
            self.providers: List[Provider] = list(map(Provider, manifest["providers"]))
            self.templates: List[PipelineTemplateMetadata] = list(map(PipelineTemplateMetadata, manifest["templates"]))
        except (FileNotFoundError, KeyError, TypeError, yaml.YAMLError) as ex:
            raise AppPipelineTemplateManifestException(
                "SAM pipeline templates manifest file is not found or ill-formatted. This could happen if the file "
                f"{manifest_path} got deleted or modified."
                "If you believe this is not the case, please file an issue at https://github.com/aws/aws-sam-cli/issues"
            ) from ex



"""
Provides methods to generate and send metrics
"""
import logging
import platform
import uuid
from dataclasses import dataclass
from functools import reduce, wraps
from pathlib import Path
from timeit import default_timer
from typing import Any, Dict, Optional

import click

from samcli import __version__ as samcli_version
from samcli.cli.context import Context
from samcli.cli.global_config import GlobalConfig
from samcli.commands._utils.experimental import get_all_experimental_statues
from samcli.commands.exceptions import UnhandledException, UserException
from samcli.lib.hook.exceptions import InvalidHookPackageConfigException
from samcli.lib.hook.hook_config import HookPackageConfig
from samcli.lib.hook.hook_wrapper import INTERNAL_PACKAGES_ROOT
from samcli.lib.hook.utils import get_hook_metadata
from samcli.lib.iac.cdk.utils import is_cdk_project
from samcli.lib.iac.plugins_interfaces import ProjectTypes
from samcli.lib.telemetry.cicd import CICDDetector, CICDPlatform
from samcli.lib.telemetry.event import EventTracker
from samcli.lib.telemetry.project_metadata import get_git_remote_origin_url, get_initial_commit_hash, get_project_name
from samcli.lib.telemetry.telemetry import Telemetry
from samcli.lib.warnings.sam_cli_warning import TemplateWarningsChecker

LOG = logging.getLogger(__name__)

WARNING_ANNOUNCEMENT = "WARNING: {}"

"""
Global variables are evil but this is a justified usage.
This creates a versatile telemetry tracking no matter where in the code. Something like a Logger.
No side effect will result in this as it is write-only for code outside of telemetry.
Decorators should be used to minimize logic involving telemetry.
"""
_METRICS = dict()


@dataclass
class ProjectDetails:
    project_type: str
    hook_name: Optional[str]
    hook_package_version: Optional[str]


def send_installed_metric():
    LOG.debug("Sending Installed Metric")

    telemetry = Telemetry()
    metric = Metric("installed")
    metric.add_data("osPlatform", platform.system())
    metric.add_data("telemetryEnabled", bool(GlobalConfig().telemetry_enabled))
    telemetry.emit(metric, force_emit=True)


def track_template_warnings(warning_names):
    """
    Decorator to track when a warning is emitted. This method accepts name of warning and executes the function,
    gathers all relevant metrics, reports the metrics and returns.

    On your warning check method use as follows

        @track_warning(['Warning1', 'Warning2'])
        def check():
            return True, 'Warning applicable'
    """

    def decorator(func):
        """
        Actual decorator method with warning names
        """

        def wrapped(*args, **kwargs):
            telemetry = Telemetry()
            template_warning_checker = TemplateWarningsChecker()
            ctx = Context.get_current_context()

            try:
                ctx.template_dict
            except AttributeError:
                LOG.debug("Ignoring warning check as template is not provided in context.")
                return func(*args, **kwargs)
            for warning_name in warning_names:
                warning_message = template_warning_checker.check_template_for_warning(warning_name, ctx.template_dict)
                metric = Metric("templateWarning")
                metric.add_data("awsProfileProvided", bool(ctx.profile))
                metric.add_data("debugFlagProvided", bool(ctx.debug))
                metric.add_data("region", ctx.region or "")
                metric.add_data("warningName", warning_name)
                metric.add_data("warningCount", 1 if warning_message else 0)  # 1-True or 0-False
                telemetry.emit(metric)

                if warning_message:
                    click.secho(WARNING_ANNOUNCEMENT.format(warning_message), fg="yellow")

            return func(*args, **kwargs)

        return wrapped

    return decorator


def track_command(func):
    """
    Decorator to track execution of a command. This method executes the function, gathers all relevant metrics,
    reports the metrics and returns.

    If you have a Click command, you can track as follows:

    .. code:: python
        @click.command(...)
        @click.options(...)
        @track_command
        def hello_command():
            print('hello')

    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        exception = None
        return_value = None
        exit_reason = "success"
        exit_code = 0

        duration_fn = _timer()

        ctx = None
        try:
            # we have get_current_context in it's own try/except to catch the RuntimeError for this and not func()
            ctx = Context.get_current_context()
        except RuntimeError:
            LOG.debug("Unable to find Click Context for getting session_id.")

        try:
            if ctx and ctx.exception:
                # re-raise here to handle exception captured in context and not run func()
                raise ctx.exception

            # Execute the function and capture return value. This is returned by the wrapper
            # First argument of all commands should be the Context
            return_value = func(*args, **kwargs)
        except (
            UserException,
            click.Abort,
            click.BadOptionUsage,
            click.BadArgumentUsage,
            click.BadParameter,
            click.UsageError,
        ) as ex:
            # Capture exception information and re-raise it later,
            # so metrics can be sent.
            exception = ex
            # NOTE(sriram-mv): Set exit code to 1 if deemed to be user fixable error.
            exit_code = 1
            if hasattr(ex, "wrapped_from") and ex.wrapped_from:
                exit_reason = ex.wrapped_from
            else:
                exit_reason = type(ex).__name__
        except Exception as ex:
            command = ctx.command_path if ctx else ""
            exception = UnhandledException(command, ex)
            # Standard Unix practice to return exit code 255 on fatal/unhandled exit.
            exit_code = 255
            exit_reason = type(ex).__name__

        if ctx:
            time = duration_fn()

            try:
                # metrics also contain a call to Context.get_current_context, catch RuntimeError
                _send_command_run_metrics(ctx, time, exit_reason, exit_code, **kwargs)
            except RuntimeError:
                LOG.debug("Unable to find Click context when sending metrics to telemetry")

        if exception:
            raise exception  # pylint: disable=raising-bad-type

        return return_value

    return wrapped


def _send_command_run_metrics(ctx: Context, duration: int, exit_reason: str, exit_code: int, **kwargs) -> None:
    """
    Emits metrics based on the results of a command run

    Parameters
    ----------
    ctx: Context
        The click context containing parameters, options, etc
    duration: int
        The total run time of the command in milliseconds
    exit_reason: str
        The exit reason from the command, "success" if successful, otherwise name of exception
    exit_code: int
        The exit code of command run
    """
    telemetry = Telemetry()

    # get_all_experimental_statues() returns Dict[str, bool]
    # since we append other values here (not just bool), need to explicitly set type
    metric_specific_attributes: Dict[str, Any] = get_all_experimental_statues() if ctx.experimental else {}

    try:
        template_dict = ctx.template_dict
        project_details = _get_project_details(kwargs.get("hook_name", ""), template_dict)
        if project_details.project_type == ProjectTypes.CDK.value:
            EventTracker.track_event("UsedFeature", "CDK")
        metric_specific_attributes["projectType"] = project_details.project_type
        if project_details.hook_name:
            metric_specific_attributes["hookPackageId"] = project_details.hook_name
        if project_details.hook_package_version:
            metric_specific_attributes["hookPackageVersion"] = project_details.hook_package_version
    except AttributeError:
        LOG.debug("Template is not provided in context, skip adding project type metric")

    metric_name = "commandRunExperimental" if ctx.experimental else "commandRun"
    metric = Metric(metric_name)
    metric.add_data("awsProfileProvided", bool(ctx.profile))
    metric.add_data("debugFlagProvided", bool(ctx.debug))
    metric.add_data("region", ctx.region or "")
    metric.add_data("commandName", ctx.command_path)  # Full command path. ex: sam local start-api

    if not ctx.command_path.endswith("init") or ctx.command_path.endswith("pipeline init"):
        # Project metadata
        # We don't capture below usage attributes for sam init as the command is not run inside a project
        metric_specific_attributes["gitOrigin"] = get_git_remote_origin_url()
        metric_specific_attributes["projectName"] = get_project_name()
        metric_specific_attributes["initialCommit"] = get_initial_commit_hash()

    metric.add_data("metricSpecificAttributes", metric_specific_attributes)
    # Metric about command's execution characteristics
    metric.add_data("duration", duration)
    metric.add_data("exitReason", exit_reason)
    metric.add_data("exitCode", exit_code)
    EventTracker.send_events()  # Sends Event metrics to Telemetry before commandRun metrics
    telemetry.emit(metric)


def _get_project_details(hook_name: str, template_dict: Dict) -> ProjectDetails:
    if not hook_name:
        hook_metadata = get_hook_metadata(template_dict)
        if not hook_metadata:
            project_type = ProjectTypes.CDK.value if is_cdk_project(template_dict) else ProjectTypes.CFN.value
            return ProjectDetails(project_type=project_type, hook_name=None, hook_package_version=None)
        hook_name = str(hook_metadata.get("HookName"))
    hook_location = Path(INTERNAL_PACKAGES_ROOT, hook_name)
    try:
        hook_package_config = HookPackageConfig(package_dir=hook_location)
    except InvalidHookPackageConfigException:
        return ProjectDetails(project_type=hook_name, hook_name=hook_name, hook_package_version=None)
    return ProjectDetails(
        project_type=hook_package_config.iac_framework,
        hook_name=hook_package_config.name,
        hook_package_version=hook_package_config.version,
    )


def _timer():
    """
    Timer to measure the elapsed time between two calls in milliseconds. When you first call this method,
    we will automatically start the timer. The return value is another method that, when called, will end the timer
    and return the duration between the two calls.

    ..code:
    >>> import time
    >>> duration_fn = _timer()
    >>> time.sleep(5)  # Say, you sleep for 5 seconds in between calls
    >>> duration_ms = duration_fn()
    >>> print(duration_ms)
        5010

    Returns
    -------
    function
        Call this method to end the timer and return duration in milliseconds

    """
    start = default_timer()

    def end():
        # time might go backwards in rare scenarios, hence the 'max'
        return int(max(default_timer() - start, 0) * 1000)  # milliseconds

    return end


def _parse_attr(obj, name):
    """
    Get attribute from an object.
    @param obj Object
    @param name Attribute name to get from the object.
        Can be nested with "." in between.
        For example: config.random_field.value
    """
    return reduce(getattr, name.split("."), obj)


def capture_parameter(metric_name, key, parameter_identifier, parameter_nested_identifier=None, as_list=False):
    """
    Decorator for capturing one parameter of the function.

    :param metric_name Name of the metric
    :param key Key for storing the captured parameter
    :param parameter_identifier Either a string for named parameter or int for positional parameter.
        "self" can be accessed with 0.
    :param parameter_nested_identifier If specified, the attribute pointed by this parameter will be stored instead.
        Can be in nested format such as config.random_field.value.
    :param as_list Default to False. Setting to True will append the captured parameter into
        a list instead of overriding the previous one.
    """

    def wrap(func):
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            return_value = func(*args, **kwargs)
            if isinstance(parameter_identifier, int):
                parameter = args[parameter_identifier]
            elif isinstance(parameter_identifier, str):
                parameter = kwargs[parameter_identifier]
            else:
                return return_value

            if parameter_nested_identifier:
                parameter = _parse_attr(parameter, parameter_nested_identifier)

            if as_list:
                add_metric_list_data(metric_name, key, parameter)
            else:
                add_metric_data(metric_name, key, parameter)
            return return_value

        return wrapped_func

    return wrap


def capture_return_value(metric_name, key, as_list=False):
    """
    Decorator for capturing the return value of the function.

    :param metric_name Name of the metric
    :param key Key for storing the captured parameter
    :param as_list Default to False. Setting to True will append the captured parameter into
        a list instead of overriding the previous one.
    """

    def wrap(func):
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            return_value = func(*args, **kwargs)
            if as_list:
                add_metric_list_data(metric_name, key, return_value)
            else:
                add_metric_data(metric_name, key, return_value)
            return return_value

        return wrapped_func

    return wrap


def add_metric_data(metric_name, key, value):
    _get_metric(metric_name).add_data(key, value)


def add_metric_list_data(metric_name, key, value):
    _get_metric(metric_name).add_list_data(key, value)


def _get_metric(metric_name):
    if metric_name not in _METRICS:
        _METRICS[metric_name] = Metric(metric_name)
    return _METRICS[metric_name]


def emit_metric(metric_name):
    if metric_name not in _METRICS:
        return
    telemetry = Telemetry()
    telemetry.emit(_get_metric(metric_name))
    _METRICS.pop(metric_name)


def emit_all_metrics():
    for key in list(_METRICS):
        emit_metric(key)


class Metric:
    """
    Metric class to store metric data and adding common attributes
    """

    def __init__(self, metric_name, should_add_common_attributes=True):
        self._data = dict()
        self._metric_name = metric_name
        self._gc = GlobalConfig()
        self._session_id = self._default_session_id()
        self._cicd_detector = CICDDetector()
        if not self._session_id:
            self._session_id = ""
        if should_add_common_attributes:
            self._add_common_metric_attributes()

    def add_list_data(self, key, value):
        if key not in self._data:
            self._data[key] = list()

        if not isinstance(self._data[key], list):
            # raise MetricDataNotList()
            return

        self._data[key].append(value)

    def add_data(self, key, value):
        self._data[key] = value

    def get_data(self):
        return self._data

    def get_metric_name(self):
        return self._metric_name

    def _add_common_metric_attributes(self):
        self._data["requestId"] = str(uuid.uuid4())
        self._data["installationId"] = self._gc.installation_id
        self._data["sessionId"] = self._session_id
        self._data["executionEnvironment"] = self._get_execution_environment()
        self._data["ci"] = bool(self._cicd_detector.platform())
        self._data["pyversion"] = platform.python_version()
        self._data["samcliVersion"] = samcli_version

    @staticmethod
    def _default_session_id() -> Optional[str]:
        """
        Get the default SessionId from Click Context.
        Fail silently if Context does not exist.
        """
        try:
            ctx = Context.get_current_context()
            if ctx:
                return ctx.session_id
            return None
        except RuntimeError:
            LOG.debug("Unable to find Click Context for getting session_id.")
            return None

    def _get_execution_environment(self) -> str:
        """
        Returns the environment in which SAM CLI is running. Possible options are:

        CLI (default)               - SAM CLI was executed from terminal or a script.
        other CICD platform name    - SAM CLI was executed in CICD

        Returns
        -------
        str
            Name of the environment where SAM CLI is executed in.
        """
        cicd_platform: Optional[CICDPlatform] = self._cicd_detector.platform()
        if cicd_platform:
            return cicd_platform.name
        return "CLI"


class MetricDataNotList(Exception):
    pass



"""
Utilities for sam deploy command
"""

import copy
import json
import textwrap

import click


def print_deploy_args(
    stack_name,
    s3_bucket,
    image_repository,
    region,
    capabilities,
    parameter_overrides,
    confirm_changeset,
    signing_profiles,
    use_changeset,
    disable_rollback,
):
    """
    Print a table of the values that are used during a sam deploy.
    Irrespective of if a image_repository is provided or not, the template is always uploaded to the s3 bucket.

    Example below:

        Deploying with following values
        ===============================
        Stack name                 : sam-app
        Region                     : us-east-1
        Confirm changeset          : False
        Disable rollback           : False
        Deployment s3 bucket       : aws-sam-cli-managed-default-samclisourcebucket-abcdef
        Capabilities               : ["CAPABILITY_IAM"]
        Parameter overrides        : {'MyParamater': '***', 'Parameter2': 'dd'}
        Signing Profiles           : {'MyFunction': 'ProfileName:ProfileOwner'}

    :param stack_name: Name of the stack used during sam deploy
    :param s3_bucket: Name of s3 bucket used for packaging code artifacts.
    :param image_repository: Name of image repository used for packaging artifacts as container images.
    :param region: Name of region to which the current sam/cloudformation stack will be deployed to.
    :param capabilities: Corresponding IAM capabilities to be used during the stack deploy.
    :param parameter_overrides: Cloudformation parameter overrides to be supplied based on the stack's template
    :param confirm_changeset: Prompt for changeset to be confirmed before going ahead with the deploy.
    :param signing_profiles: Signing profile details which will be used to sign functions/layers
    :param use_changeset: Flag to use or skip the usage of changesets
    :param disable_rollback: Preserve the state of previously provisioned resources when an operation fails.
    """
    _parameters = parameter_overrides.copy()

    capabilities_string = json.dumps(capabilities)

    _signing_profiles = {}
    if signing_profiles:
        for key, value in signing_profiles.items():
            _signing_profiles[key] = f"{value['profile_name']}:{value['profile_owner']}"

    image_repository_format_text = (
        json.dumps(image_repository, indent=4) if isinstance(image_repository, dict) else image_repository
    )
    parameter_overrides_format_text = json.dumps(_parameters)
    signing_profiles_format_text = json.dumps(signing_profiles)

    click.secho("\n\tDeploying with following values\n\t===============================", fg="yellow")
    click.echo(f"\tStack name                   : {stack_name}")
    click.echo(f"\tRegion                       : {region}")
    if use_changeset:
        click.echo(f"\tConfirm changeset            : {confirm_changeset}")
    click.echo(f"\tDisable rollback             : {disable_rollback}")
    if image_repository:
        msg = "Deployment image repository  : "
        # NOTE(sriram-mv): tab length is 8 spaces.
        prefix_length = len(msg) + 8
        click.echo(f"\t{msg}")
        click.echo(f"{textwrap.indent(image_repository_format_text, prefix=' ' * prefix_length)}")
    click.echo(f"\tDeployment s3 bucket         : {s3_bucket}")
    click.echo(f"\tCapabilities                 : {capabilities_string}")
    click.echo(f"\tParameter overrides          : {parameter_overrides_format_text}")
    click.echo(f"\tSigning Profiles             : {signing_profiles_format_text}")

    click.secho("\nInitiating deployment\n=====================\n", fg="yellow")


def sanitize_parameter_overrides(parameter_overrides):
    """
    Get sanitized parameter override values based on if the workflow went via a guided deploy to set the
    parameter overrides for deployment. If a guided deploy was followed the parameter overrides consists
    of additional information such as if a given parameter's value is hidden or not.
    :param parameter_overrides: dictionary of parameter key values.
    :return:
    """
    return {key: value.get("Value") if isinstance(value, dict) else value for key, value in parameter_overrides.items()}


def hide_noecho_parameter_overrides(template_parameters, parameter_overrides):
    hidden_params = copy.deepcopy(parameter_overrides)
    params = template_parameters.get("Parameters", None)
    for key, value in hidden_params.items():
        if isinstance(params, dict) and key in params and isinstance(params[key], dict):
            is_hidden = params[key].get("NoEcho", False)
            hidden_params[key] = value if not is_hidden else "*" * 5
    return hidden_params



"""
This is the core module of the cookiecutter workflow, it defines how to create a template, prompt the user for
values of the context and how to generate a project from the given template and provided context
"""
import logging
from typing import Dict, List, Optional

from cookiecutter.exceptions import RepositoryNotFound, UnknownRepoType
from cookiecutter.main import cookiecutter

from samcli.commands.exceptions import UserException
from samcli.lib.init.arbitrary_project import generate_non_cookiecutter_project

from .exceptions import GenerateProjectFailedError, InvalidLocationError, PostprocessingError, PreprocessingError
from .interactive_flow import InteractiveFlow
from .plugin import Plugin
from .processor import Processor

LOG = logging.getLogger(__name__)


class Template:
    """
    A class for project generation based on cookiecutter template

    Attributes
    ----------
    location: str
        Template location (git, mercurial, http(s), path)
    interactive_flows: Optional[List[InteractiveFlow]]
        An optional series of interactive questions to be asked to the user where the answers are used to fulfill
        the values of the cookiecutter.json keys, i.e. the cookiecutter context
    preprocessors: Optional[List[Processor]]
        An optional series of processes to be executed before generating the cookiecutter project. These processes
        are usually used to manipulate the cookiecutter context or doing some bootstrapping before generating
        the cookiecutter project.
        preprocessors are different than cookiecutter's hooks in the sense that they can mutate the context while
        the pre_gen_project hooks can't.
    postprocessors: Optional[List[Processor]]
        An optional series of processes to be executed after generating the cookiecutter project. These processes
        are usually used to clean up resources and to print messages to the user after successfully generating
        the cookiecutter project.
    plugins: Optional[List[Plugin]]
        An optional series of plugins to be plugged in. A plugin defines its own interactive_flow, preprocessor
        and postprocessor. A plugin is a sub-set of the template, if there is a common behavior among multiple
        templates, it is better to be extracted to a plugin that can then be plugged in to each of these templates.
    metadata: Optional[Dict]
        An optional dictionary with extra information about the template

    Methods
    -------
    run_interactive_flows() -> Dict:
        runs the interactive_flows(questions) one be one then combines and returns the answers in one dictionary
        in the form of {question.key: answer}
    generate_project(context) -> None:
        runs the preprocessors chain passing in the given context to the first processor which process it then produce
        a new context which is passed to the next preprocessor and so on.
        After that, it execute the cookiecutter generator to generate the project with the final preprocessed context.
        Finally, it runs the postprocessors chain with the same manner like the preprocessors.
    """

    def __init__(
        self,
        location: str,
        interactive_flows: Optional[List[InteractiveFlow]] = None,
        preprocessors: Optional[List[Processor]] = None,
        postprocessors: Optional[List[Processor]] = None,
        plugins: Optional[List[Plugin]] = None,
        metadata: Optional[Dict] = None,
    ):
        """
        Initialize the class

        Parameters
        ----------
        location: str
            Template location (git, mercurial, http(s), path)
        interactive_flows: Optional[List[InteractiveFlow]]
            An optional series of interactive questions to be asked to the user where the answers are used to fulfill
            the values of the cookiecutter.json keys, i.e. the cookiecutter context
        preprocessors: Optional[List[Processor]]
            An optional series of processes to be executed before generating the cookiecutter project. These processes
            are usually used to manipulate the cookiecutter context or doing some bootstrapping before generating
            the cookiecutter project.
        postprocessors: Optional[List[Processor]]
            An optional series of processes to be executed after generating the cookiecutter project. These processes
            are usually used to clean up resources and to print messages to the user after successfully generating
            the cookiecutter project.
        plugins: Optional[List[Plugin]]
            An optional series of plugins to be plugged in. A plugin defines its own interactive_flow, preprocessor
            and postprocessor. A plugin is a sub-set of the template, if there is a common behavior among multiple
            templates, it is better to be extracted to a plugin that can then be plugged in to each of these templates.
        metadata: Optional[Dict]
            An optional dictionary with extra information about the template
        """
        self._location = location
        self._interactive_flows = interactive_flows or []
        self._preprocessors = preprocessors or []
        self._postprocessors = postprocessors or []
        self._plugins = plugins or []
        for plugin in self._plugins:
            if plugin.interactive_flow:
                self._interactive_flows.append(plugin.interactive_flow)
            if plugin.preprocessor:
                self._preprocessors.append(plugin.preprocessor)
            if plugin.postprocessor:
                self._postprocessors.append(plugin.postprocessor)
        self.metadata = metadata

    def run_interactive_flows(self, context: Optional[Dict] = None) -> Dict:
        """
        prompt the user a series of questions' flows and gather the answers to create the cookiecutter context.
        The questions are identified by keys. If multiple questions, whether within the same flow or across
        different flows, have the same key, the last question will override the others and we will get only the
        answer of this question.

        Raises:
            UserException(ClickException) if anything went wrong.

        Returns:
            A Dictionary in the form of {question.key: answer} representing user's answers to the flows' questions
        """
        try:
            context = context if context else {}
            context["shared_values"] = "default"
            for flow in self._interactive_flows:
                context = flow.run(context)
            return context
        except Exception as e:
            raise UserException(str(e), wrapped_from=e.__class__.__name__) from e

    def generate_project(self, context: Dict, output_dir: str) -> None:
        """
        Generates a project based on this cookiecutter template and the given context. The context is first
        processed and manipulated by series of preprocessors(if any) then the project is generated and finally
        the series of postprocessors(if any) are executed.

        Parameters
        ----------
        context: Dict
            the cookiecutter context to fulfill the values of cookiecutter.json keys
        output_dir: str
            the directory where project will be generated in

        Raises
        ------
        PreprocessingError
            xxx
        InvalidLocationError
            if the given location is not from a known repo type
        GenerateProjectFailedError
            if something else went wrong
        PostprocessingError
            yyy
        """
        try:
            LOG.debug("preprocessing the cookiecutter context")
            for processor in self._preprocessors:
                context = processor.run(context)
        except Exception as e:
            raise PreprocessingError(template=self._location, provider_error=e) from e

        try:
            LOG.debug("Baking a new template with cookiecutter with all parameters")
            cookiecutter(
                template=self._location,
                output_dir=output_dir,
                no_input=True,
                extra_context=context,
                overwrite_if_exists=True,
            )
        except RepositoryNotFound:
            # cookiecutter.json is not found in the template. Let's just clone it directly without
            # using cookiecutter and call it done.
            LOG.debug(
                "Unable to find cookiecutter.json in the project. Downloading it directly without treating "
                "it as a cookiecutter template"
            )
            generate_non_cookiecutter_project(location=self._location, output_dir=".")
        except UnknownRepoType as e:
            raise InvalidLocationError(template=self._location) from e
        except Exception as e:
            raise GenerateProjectFailedError(template=self._location, provider_error=e) from e

        try:
            LOG.debug("postprocessing the cookiecutter context")
            for processor in self._postprocessors:
                context = processor.run(context)
        except Exception as e:
            raise PostprocessingError(template=self._location, provider_error=e) from e



"""
Hooks Wrapper Class
"""
import importlib
import logging
from pathlib import Path
from typing import Dict, List, Optional, cast

from .exceptions import (
    HookPackageExecuteFunctionalityException,
    InvalidHookWrapperException,
)
from .hook_config import HookPackageConfig

LOG = logging.getLogger(__name__)
INTERNAL_PACKAGES_ROOT = Path(__file__).parent / ".." / ".." / "hook_packages"


class IacHookWrapper:
    """IacHookWrapper

    An IacHookWrapper instance, upon instantiation, looks up the hook package with the specified hook package ID.
    It provides the "prepare" method, which generates an IaC metadata and output the location of the metadata file.

    Example:
    ```
    hook = IacHookWrapper("terraform")
    metadata_loc = hook.prepare("path/to/iac_project", "path/to/output", True)
    ```
    """

    _hook_name: str
    _config: Optional[HookPackageConfig]

    def __init__(self, hook_name: str):
        """
        Parameters
        ----------
        hook_name: str
            Hook name
        """
        self._hook_name = hook_name
        self._config = None
        self._load_hook_package(hook_name)

    def prepare(
        self,
        output_dir_path: str,
        iac_project_path: Optional[str] = None,
        debug: bool = False,
        aws_profile: Optional[str] = None,
        aws_region: Optional[str] = None,
        skip_prepare_infra: bool = False,
    ) -> str:
        """
        Run the prepare hook to generate the IaC Metadata file.

        Parameters
        ----------
        output_dir_path: str
            the path where the hook can create the generated Metadata files. Required
        iac_project_path: str
            the path where the hook can find the TF application. Default value in current work directory.
        debug: bool
            True/False flag to tell the hooks if should print debugging logs or not. Default is False.
        aws_profile: str
            AWS profile to use. Default is None (use default profile)
        aws_region: str
            AWS region to use. Default is None (use default region)
        skip_prepare_infra: bool
            Flag to skip skip prepare hook if we already have the metadata file. Default is False.

        Returns
        -------
        str
            Path to the generated IaC Metadata file
        """
        LOG.info('Executing prepare hook of hook "%s"', self._hook_name)
        params = {
            "IACProjectPath": iac_project_path if iac_project_path else str(Path.cwd()),
            "OutputDirPath": output_dir_path,
            "Debug": debug,
            "SkipPrepareInfra": skip_prepare_infra,
        }
        if aws_profile:
            params["Profile"] = aws_profile
        if aws_region:
            params["Region"] = aws_region

        output = self._execute("prepare", params)

        metadata_file_loc = None
        iac_applications: Dict[str, Dict] = output.get("iac_applications", {})
        if iac_applications and len(iac_applications) == 1:
            # NOTE: we assume there is only one application in the `iac_applications` dictionary,
            # which is the only case we support right now
            main_application = list(iac_applications.values())[0]
            metadata_file_loc = main_application.get("metadata_file")

        if not metadata_file_loc:
            raise InvalidHookWrapperException("Metadata file path not found in the prepare hook output")

        LOG.debug("Metadata file location - %s", metadata_file_loc)
        return cast(str, metadata_file_loc)

    def _load_hook_package(self, hook_name: str) -> None:
        """Find and load hook package config with given hook name

        Parameters
        ----------
        hook_name: str
            Hook name
        """
        # locate hook package from internal first
        LOG.debug("Looking for internal hook package")
        for child in INTERNAL_PACKAGES_ROOT.iterdir():
            if child.name == hook_name:
                LOG.debug('Loaded internal hook package "%s"', hook_name)
                self._config = HookPackageConfig(child)
                return

        raise InvalidHookWrapperException(f'Cannot locate hook package with hook_name "{hook_name}"')

    def _execute(self, functionality_key: str, params: Optional[Dict] = None) -> Dict:
        """
        Execute a functionality with given key

        Parameters
        ----------
        functionality_key: str
            The key of the functionality
        params: Dict
            A dict of parameters to pass into the execution

        Returns
        -------
        Dict
            the output from the execution
        """
        if not self._config:
            raise InvalidHookWrapperException("Config is missing. You must instantiate a hook with a valid config")

        if functionality_key not in self._config.functionalities:  # pylint: disable=unsupported-membership-test
            raise HookPackageExecuteFunctionalityException(
                f'Functionality "{functionality_key}" is not defined in the hook package'
            )

        functionality = self._config.functionalities[functionality_key]
        if functionality.entry_method:
            return _execute_as_module(functionality.module, functionality.method, params)

        raise InvalidHookWrapperException(f'Functionality "{functionality_key}" is missing an "entry_method"')


def _execute_as_module(module: str, method: str, params: Optional[Dict] = None) -> Dict:
    """
    Execute a module/method with given module and given method

    Parameters
    ----------
    module: str
        the module where the method lives in
    method: str
        the name of the method to execute
    params: Dict
        A dict of parameters to pass into the execution

    Returns
    -------
    Dict
        the output from the execution
    """
    try:
        mod = importlib.import_module(module)
    except ImportError as e:
        raise InvalidHookWrapperException(f'Import error - HookFunctionality module "{module}"') from e

    if not hasattr(mod, method):
        raise InvalidHookWrapperException(f'HookFunctionality module "{module}" has no method "{method}"')

    result = getattr(mod, method)(params)
    return cast(Dict, result)


def get_available_hook_packages_ids() -> List[str]:
    """
    return a list of available hook names.

    Returns
    -------
    List
        The available hook names.
    """
    LOG.debug("Return available internal hook packages")
    hook_packages_ids = []
    for child in INTERNAL_PACKAGES_ROOT.iterdir():
        if child.is_dir() and child.name[0].isalpha():
            hook_packages_ids.append(child.name)

    return hook_packages_ids



""" This module represents the questions to ask to the user to fulfill the cookiecutter context. """
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

import click


class QuestionKind(Enum):
    """An Enum of possible question types."""

    info = "info"
    confirm = "confirm"
    choice = "choice"
    question = "question"


class Promptable(ABC):
    """
    Abstract class Question, Info, Choice, Confirm implement.
    These classes need to implement their own prompt() method to prompt differently.
    """

    @abstractmethod
    def prompt(self, text: str, default_answer: Optional[Any]) -> Any:
        pass


class Question(Promptable):
    """
    A question to be prompt to the user in an interactive flow where the response is used to fulfill
    the cookiecutter context.

    Attributes
    ----------
    _key: str
        The key of the cookiecutter config to associate the answer with.
    _text: str
        The text to prompt to the user
    _required: bool
        Whether the user must provide an answer for this question or not.
    _default_answer: Optional[Union[str, Dict]]
        A default answer that is suggested to the user,
        it can be directly provided (a string)
        or resolved from cookiecutter context (a Dict, in the form of {"keyPath": [...,]})
    _next_question_map: Optional[Dict[str, str]]
        A simple branching mechanism, it refers to what is the next question to ask the user if he answered
        a particular answer to this question. this map is in the form of {answer: next-question-key}. this
        can be used in a flow like:
            supposed we are prompting the user for two resources r1 & r2 such that if the resource already
            exist then the user provide it otherwise we created it on his behave. the questions may look like:
                Question(key="r1" text="do you already have resource1?",
                         _next_question_map={"True": "r2", "False": "r1-creation"}
                Question(key="r1-creation" text="What do you like to name resource1?"}
                Question(key="r2" text="do you already have resource2?",
                         _next_question_map={"True": "r3", "False": "r2-creation"}
    _default_next_question_key: str
        The key of the next question that is not based on user's answer, i.e., if there is no matching in
        the _next_question_map
    """

    def __init__(
        self,
        key: str,
        text: str,
        default: Optional[Union[str, Dict]] = None,
        is_required: Optional[bool] = None,
        allow_autofill: Optional[bool] = None,
        next_question_map: Optional[Dict[str, str]] = None,
        default_next_question_key: Optional[str] = None,
    ):
        self._key = key
        self._text = text
        self._required = is_required
        self._allow_autofill = allow_autofill
        self._default_answer = default
        # if it is an optional question, set an empty default answer to prevent click from keep asking for an answer
        if not self._required and self._default_answer is None:
            self._default_answer = ""
        self._next_question_map: Dict[str, str] = next_question_map or {}
        self._default_next_question_key = default_next_question_key

    @property
    def key(self) -> str:
        return self._key

    @property
    def text(self) -> str:
        return self._text

    @property
    def default_answer(self) -> Optional[Any]:
        return self._resolve_default_answer()

    @property
    def required(self) -> Optional[bool]:
        return self._required

    @property
    def next_question_map(self) -> Dict[str, str]:
        return self._next_question_map

    @property
    def default_next_question_key(self) -> Optional[str]:
        return self._default_next_question_key

    def ask(self, context: Optional[Dict] = None) -> Any:
        """
        prompt the user this question

        Parameters
        ----------
        context
            The cookiecutter context dictionary containing previous questions' answers and default values

        Returns
        -------
        The user provided answer.
        """
        resolved_default_answer = self._resolve_default_answer(context)

        # skip the question and directly use the default value if autofill is allowed.
        if resolved_default_answer is not None and self._allow_autofill:
            return resolved_default_answer

        # if it is an optional question with no default answer,
        # set an empty default answer to prevent click from keep asking for an answer
        if not self._required and resolved_default_answer is None:
            resolved_default_answer = ""

        return self.prompt(self._resolve_text(context), resolved_default_answer)

    def prompt(self, text: str, default_answer: Optional[Any]) -> Any:
        return click.prompt(text=text, default=default_answer)

    def get_next_question_key(self, answer: Any) -> Optional[str]:
        # _next_question_map is a Dict[str(answer), str(next question key)]
        # so we need to parse the passed argument(answer) to str
        answer = str(answer)
        return self._next_question_map.get(answer, self._default_next_question_key)

    def set_default_next_question_key(self, next_question_key: str) -> None:
        self._default_next_question_key = next_question_key

    def _resolve_key_path(self, key_path: List, context: Dict) -> List[str]:
        """
        key_path element is a list of str and Dict.
        When the element is a dict, in the form of { "valueOf": question_key },
        it means it refers to the answer to another questions.
        _resolve_key_path() will replace such dict with the actual question answer

        Parameters
        ----------
        key_path
            The key_path list containing str and dict
        context
            The cookiecutter context containing answers to previous answered questions
        Returns
        -------
            The key_path list containing only str
        """
        resolved_key_path: List[str] = []
        for unresolved_key in key_path:
            if isinstance(unresolved_key, str):
                resolved_key_path.append(unresolved_key)
            elif isinstance(unresolved_key, dict):
                if "valueOf" not in unresolved_key:
                    raise KeyError(f'Missing key "valueOf" in question default keyPath element "{unresolved_key}".')
                query_question_key: str = unresolved_key.get("valueOf", "")
                if query_question_key not in context:
                    raise KeyError(
                        f'Invalid question key "{query_question_key}" referenced '
                        f"in default answer of question {self.key}"
                    )
                resolved_key_path.append(context[query_question_key])
            else:
                raise ValueError(f'Invalid value "{unresolved_key}" in key path')
        return resolved_key_path

    def _resolve_value_from_expression(self, expression: Any, context: Optional[Dict] = None) -> Optional[Any]:
        """
        a question may have a value provided directly as string or number value
        or indirectly from cookiecutter context using a key path

        Parameters
        ----------
        context
            Cookiecutter context used to resolve values.

        Raises
        ------
        KeyError
            When an expression depends on the answer to a non-existent question
        ValueError
            The expression is malformed

        Returns
        -------
        Optional value, it might be resolved from cookiecutter context using specified key path.

        """
        if isinstance(expression, dict):
            context = context if context else {}

            # load value using key path from cookiecutter
            if "keyPath" not in expression:
                raise KeyError(f'Missing key "keyPath" in "{expression}".')
            unresolved_key_path = expression.get("keyPath", [])
            if not isinstance(unresolved_key_path, list):
                raise ValueError(f'Invalid expression "{expression}" in question {self.key}')

            return context.get(str(self._resolve_key_path(unresolved_key_path, context)))
        return expression

    def _resolve_text(self, context: Optional[Dict] = None) -> str:
        resolved_text = self._resolve_value_from_expression(self._text, context)
        if resolved_text is None:
            raise ValueError(f"Cannot resolve value from expression: {self._text}")
        return str(resolved_text)

    def _resolve_default_answer(self, context: Optional[Dict] = None) -> Optional[Any]:
        return self._resolve_value_from_expression(self._default_answer, context)


class Info(Question):
    def prompt(self, text: str, default_answer: Optional[Any]) -> Any:
        return click.echo(message=text)


class Confirm(Question):
    def prompt(self, text: str, default_answer: Optional[Any]) -> Any:
        return click.confirm(text=text)


class Choice(Question):
    def __init__(
        self,
        key: str,
        text: str,
        options: List[str],
        default: Optional[str] = None,
        is_required: Optional[bool] = None,
        allow_autofill: Optional[bool] = None,
        next_question_map: Optional[Dict[str, str]] = None,
        default_next_question_key: Optional[str] = None,
    ):
        if not options:
            raise ValueError("No defined options")
        self._options = options
        super().__init__(key, text, default, is_required, allow_autofill, next_question_map, default_next_question_key)

    def prompt(self, text: str, default_answer: Optional[Any]) -> Any:
        click.echo(text)
        for index, option in enumerate(self._options):
            click.echo(f"\t{index + 1} - {option}")
        options_indexes = self._get_options_indexes(base=1)
        choices = list(map(str, options_indexes))
        choice = click.prompt(
            text="Choice",
            default=default_answer,
            show_choices=False,
            type=click.Choice(choices),
            show_default=default_answer is not None,
        )
        return self._options[int(choice) - 1]

    def _get_options_indexes(self, base: int = 0) -> List[int]:
        return list(range(base, len(self._options) + base))


class QuestionFactory:
    question_classes: Dict[QuestionKind, Type[Question]] = {
        QuestionKind.info: Info,
        QuestionKind.choice: Choice,
        QuestionKind.confirm: Confirm,
        QuestionKind.question: Question,
    }

    @staticmethod
    def create_question_from_json(question_json: Dict) -> Question:
        key = question_json["key"]
        text = question_json["question"]
        options = question_json.get("options")
        default = question_json.get("default")
        is_required = question_json.get("isRequired")
        allow_autofill = question_json.get("allowAutofill")
        next_question_map = question_json.get("nextQuestion")
        default_next_question = question_json.get("defaultNextQuestion")
        kind_str = question_json.get("kind")
        kind = QuestionKind(kind_str) if kind_str else QuestionKind.choice if options else QuestionKind.question

        klass = QuestionFactory.question_classes[kind]
        args = {
            "key": key,
            "text": text,
            "default": default,
            "is_required": is_required,
            "allow_autofill": allow_autofill,
            "next_question_map": next_question_map,
            "default_next_question_key": default_next_question,
        }
        if klass == Choice:
            args["options"] = options

        return klass(**args)



"""
Utils for gathering system related info, mainly for use in `sam --info`
"""

import json
from typing import Dict, cast


def gather_system_info() -> Dict[str, str]:
    """
    Gather system info

    Returns
    -------
    dict[str, str]
    """
    from platform import platform, python_version

    info = {
        "python": python_version(),
        "os": platform(),
    }
    return info


def gather_additional_dependencies_info() -> Dict[str, str]:
    """
    Gather additional depedencies info

    Returns
    -------
    dict[str, str]
        A dictionary with the key representing the info we need
        and value being the version number
    """
    info = {
        "docker_engine": _gather_docker_info(),
        "aws_cdk": _gather_cdk_info(),
        "terraform": _gather_terraform_info(),
    }
    return info


def _gather_docker_info() -> str:
    """
    Get Docker Engine version

    Returns
    -------
    str
        Version number of Docker Engine if available. Otherwise "Not available"
    """
    import contextlib

    import docker

    from samcli.local.docker.utils import is_docker_reachable

    with contextlib.closing(docker.from_env()) as client:
        if is_docker_reachable(client):
            return cast(str, client.version().get("Version", "Not available"))
        return "Not available"


def _gather_cdk_info():
    """
    Get AWS CDK version numbner.

    AWS SAM CLI does not invoke CDK, but only uses the CFN templates generated from `cdk synth`.
    Knowing the CDK version number will help us diagnose if any issue is caused by a new version of CDK.

    Returns
    -------
    str
        Version number of AWS CDK if available, e.g. "2.20.0 (build 738ef49)"
        Otherwise "Not available"
    """
    import subprocess

    try:
        process = subprocess.run(["cdk", "--version"], capture_output=True, text=True, check=True)
        return process.stdout.strip()
    except Exception:
        return "Not available"


def _gather_terraform_info():
    """
    Get Terraform version numbner.

    AWS SAM CLI invokes Terraform for Terraform applications.
    Knowing the Terraform version number will help us diagnose if any issue is caused by a new version of Terraform.

    Returns
    -------
    str
        Version number of Terraform if available, e.g. "1.2.1"
        Otherwise "Not available"
    """
    import subprocess

    try:
        process = subprocess.run(["terraform", "version", "-json"], capture_output=True, text=True, check=True)
        info_dict = json.loads(process.stdout)
        return info_dict.get("terraform_version", "Not available")
    except Exception:
        return "Not available"
